from __future__ import annotations

from typing import Dict

from netsquid.components.component import Component, Port
from netsquid.nodes import Node

from qoala.lang.ehi import EhiNetworkInfo


class NetstackComponent(Component):
    """NetSquid component representing the network stack in QNodeOS.

    Subcomponent of a QnosComponent.

    Has communications ports with
     - the processor component of this QNodeOS
     - the netstack component of the remote node
        NOTE: at this moment only a single other node is supported in the network

    This is a static container for network-stack-related components and ports.
    Behavior of a QNodeOS network stack is modeled in the `NetProcNode` class,
    which is a subclass of `Protocol`.
    """

    def __init__(self, node: Node, ehi_network: EhiNetworkInfo) -> None:
        super().__init__(f"{node.name}_netstack")
        self._node = node

        self._peer_in_ports: Dict[str, str] = {}  # peer name -> port name
        self._peer_out_ports: Dict[str, str] = {}  # peer name -> port name

        for other_node in ehi_network.nodes.values():
            if other_node == node.name:
                continue
            port_in_name = f"peer_{other_node}_in"
            port_out_name = f"peer_{other_node}_out"
            self._peer_in_ports[other_node] = port_in_name
            self._peer_out_ports[other_node] = port_out_name

        self.add_ports(self._peer_in_ports.values())
        self.add_ports(self._peer_out_ports.values())

        self.add_ports(["host_out", "host_in"])
        self.add_ports(["qnos_out", "qnos_in"])

        # Separate channel for "memory freed" signals
        self.add_ports(["qnos_mem_out", "qnos_mem_in"])

        self.add_ports(["entdist_out", "entdist_in"])

    @property
    def node_name(self) -> str:
        return self._node.name  # type: ignore

    @property
    def node_id(self) -> int:
        return self._node.ID  # type: ignore

    @property
    def qnos_in_port(self) -> Port:
        return self.ports["qnos_in"]

    @property
    def qnos_out_port(self) -> Port:
        return self.ports["qnos_out"]

    @property
    def host_in_port(self) -> Port:
        return self.ports["host_in"]

    @property
    def host_out_port(self) -> Port:
        return self.ports["host_out"]

    @property
    def qnos_mem_in_port(self) -> Port:
        return self.ports["qnos_mem_in"]

    @property
    def qnos_mem_out_port(self) -> Port:
        return self.ports["qnos_mem_out"]

    @property
    def entdist_in_port(self) -> Port:
        return self.ports["entdist_in"]

    @property
    def entdist_out_port(self) -> Port:
        return self.ports["entdist_out"]

    def peer_in_port(self, name: str) -> Port:
        port_name = self._peer_in_ports[name]
        return self.ports[port_name]

    def peer_out_port(self, name: str) -> Port:
        port_name = self._peer_out_ports[name]
        return self.ports[port_name]

    @property
    def node(self) -> Node:
        return self._node



from __future__ import annotations

from typing import Generator

from pydynaa import EventExpression
from qoala.runtime.message import Message
from qoala.sim.host.hostinterface import HostInterface


class ClassicalSocket:
    """Wrapper around classical ports"""

    def __init__(
        self, host: HostInterface, remote_name: str, pid: int, remote_pid: int
    ):
        self._host = host
        self._remote_name = remote_name
        self._pid = pid
        self._remote_pid = remote_pid

    @property
    def remote_name(self) -> str:
        return self._remote_name

    @property
    def remote_pid(self) -> int:
        return self._remote_pid

    def send(self, msg: Message) -> None:
        """Send a message to the remote node."""
        self._host.send_peer_msg(self._remote_name, msg)

    def recv(self) -> Generator[EventExpression, None, Message]:
        msg = yield from self._host.receive_peer_msg(self._remote_name)
        return msg

    def read(self) -> Message:
        return self._host.pop_msg(
            self._remote_name, dst_pid=self._pid, src_pid=self._remote_pid
        )

    def send_str(self, msg: str) -> None:
        self.send(Message(src_pid=self._pid, dst_pid=self._remote_pid, content=msg))

    def recv_str(self) -> Generator[EventExpression, None, str]:
        msg = yield from self.recv()
        assert isinstance(msg.content, str)
        return msg.content

    def read_str(self) -> str:
        msg = self.read()
        assert isinstance(msg.content, str)
        return msg.content

    def send_int(self, value: int) -> None:
        self.send(Message(src_pid=self._pid, dst_pid=self._remote_pid, content=value))

    def recv_int(self) -> Generator[EventExpression, None, int]:
        msg = yield from self.recv()
        assert isinstance(msg.content, int)
        return msg.content

    def read_int(self) -> int:
        msg = self.read()
        assert isinstance(msg.content, int)
        return msg.content

    def send_float(self, value: float) -> None:
        self.send(Message(src_pid=self._pid, dst_pid=self._remote_pid, content=value))

    def recv_float(self) -> Generator[EventExpression, None, float]:
        msg = yield from self.recv()
        assert isinstance(msg.content, float)
        return msg.content

    def read_float(self) -> float:
        msg = self.read()
        assert isinstance(msg.content, float)
        return msg.content



from __future__ import annotations

import copy
import itertools
from dataclasses import dataclass
from math import ceil, floor
from typing import Dict, FrozenSet, List, Optional, Tuple, Type

from netqasm.lang.instr.base import NetQASMInstruction
from netqasm.lang.instr.flavour import Flavour

from qoala.lang.common import MultiQubit


@dataclass(frozen=True)
class EhiQubitInfo:
    is_communication: bool
    decoherence_rate: float  # rate per second


@dataclass(frozen=True)
class EhiGateInfo:
    instruction: Type[NetQASMInstruction]
    duration: float  # ns
    decoherence: float  # rate per second, for all qubits


@dataclass(frozen=True)
class EhiLinkInfo:
    duration: float  # ns
    fidelity: float


@dataclass(frozen=True)
class EhiLatencies:
    host_instr_time: float  # duration of classical Host instr execution (CL)
    qnos_instr_time: float  # duration of classical Qnos instr execution (QL)
    host_peer_latency: float  # processing time for Host messages from remote node (CC)
    internal_sched_latency: float  # processing time for messaging between node scheduler and processor schedulers

    @classmethod
    def all_zero(cls) -> EhiLatencies:
        return EhiLatencies(0, 0, 0, 0)


@dataclass(frozen=True)
class EhiNodeInfo:
    """Hardware made available to offline compiler."""

    qubit_infos: Dict[int, EhiQubitInfo]  # qubit ID -> info

    flavour: Type[
        Flavour
    ]  # set of NetQASM instrs, no info about which qubits can do what instr
    single_gate_infos: Dict[int, List[EhiGateInfo]]  # qubit ID -> gates
    multi_gate_infos: Dict[
        MultiQubit, List[EhiGateInfo]
    ]  # ordered qubit ID list -> gates
    latencies: EhiLatencies
    all_qubit_gate_infos: Optional[
        List[EhiGateInfo]
    ] = None  # gates that are applied to all qubits

    def find_single_gate(
        self, qubit_id: int, instr: Type[NetQASMInstruction]
    ) -> Optional[EhiGateInfo]:
        if qubit_id not in self.single_gate_infos:
            return None
        for info in self.single_gate_infos[qubit_id]:
            if info.instruction == instr:
                return info
        return None

    def find_multi_gate(
        self, qubit_ids: List[int], instr: Type[NetQASMInstruction]
    ) -> Optional[EhiGateInfo]:
        multi = MultiQubit(qubit_ids)
        if multi not in self.multi_gate_infos:
            return None
        for info in self.multi_gate_infos[multi]:
            if info.instruction == instr:
                return info
        return None

    def find_all_qubit_gate(
        self, instr: Type[NetQASMInstruction]
    ) -> Optional[EhiGateInfo]:
        if self.all_qubit_gate_infos is None:
            return None

        for info in self.all_qubit_gate_infos:
            if info.instruction == instr:
                return info
        return None


class EhiBuilder:
    @classmethod
    def decoherence_qubit(
        cls, is_communication: bool, decoherence_rate: float
    ) -> EhiQubitInfo:
        return EhiQubitInfo(
            is_communication=is_communication, decoherence_rate=decoherence_rate
        )

    @classmethod
    def perfect_qubit(cls, is_communication: bool) -> EhiQubitInfo:
        return cls.decoherence_qubit(
            is_communication=is_communication, decoherence_rate=0
        )

    @classmethod
    def decoherence_gates(
        cls,
        duration: float,
        instructions: List[Type[NetQASMInstruction]],
        decoherence: float,
    ) -> List[EhiGateInfo]:
        return [
            EhiGateInfo(instruction=instr, duration=duration, decoherence=decoherence)
            for instr in instructions
        ]

    @classmethod
    def perfect_gates(
        cls, duration: float, instructions: List[Type[NetQASMInstruction]]
    ) -> List[EhiGateInfo]:
        return cls.decoherence_gates(
            duration=duration, instructions=instructions, decoherence=0
        )

    @classmethod
    def perfect_uniform(
        cls,
        num_qubits,
        flavour: Type[Flavour],
        single_instructions: List[Type[NetQASMInstruction]],
        single_duration: float,
        two_instructions: List[Type[NetQASMInstruction]],
        two_duration: float,
        all_qubit_instructions: List[Type[NetQASMInstruction]] = None,
        all_qubit_duration: float = 0,
        latencies: Optional[EhiLatencies] = None,
    ) -> EhiNodeInfo:
        if all_qubit_instructions is None:
            return cls.fully_uniform(
                num_qubits=num_qubits,
                flavour=flavour,
                qubit_info=cls.perfect_qubit(is_communication=True),
                single_gate_infos=cls.perfect_gates(
                    single_duration, single_instructions
                ),
                two_gate_infos=cls.perfect_gates(two_duration, two_instructions),
                latencies=latencies,
            )
        else:
            all_qubit_gate_infos = cls.perfect_gates(
                all_qubit_duration, all_qubit_instructions
            )
            return cls.fully_uniform(
                num_qubits=num_qubits,
                flavour=flavour,
                qubit_info=cls.perfect_qubit(is_communication=True),
                single_gate_infos=cls.perfect_gates(
                    single_duration, single_instructions
                ),
                two_gate_infos=cls.perfect_gates(two_duration, two_instructions),
                all_qubit_gate_infos=all_qubit_gate_infos,
                latencies=latencies,
            )

    @classmethod
    def fully_uniform(
        cls,
        num_qubits,
        qubit_info: EhiQubitInfo,
        flavour: Type[Flavour],
        single_gate_infos: List[EhiGateInfo],
        two_gate_infos: List[EhiGateInfo],
        all_qubit_gate_infos: List[EhiGateInfo] = None,
        latencies: Optional[EhiLatencies] = None,
    ) -> EhiNodeInfo:
        q_infos = {i: qubit_info for i in range(num_qubits)}
        sg_infos = {i: single_gate_infos for i in range(num_qubits)}
        mg_infos = {}
        for i in range(num_qubits):
            for j in range(num_qubits):
                if i != j:
                    multi = MultiQubit([i, j])
                    mg_infos[multi] = two_gate_infos

        if latencies is None:
            latencies = EhiLatencies.all_zero()

        return EhiNodeInfo(
            q_infos,
            flavour,
            sg_infos,
            mg_infos,
            latencies,
            all_qubit_gate_infos=copy.deepcopy(all_qubit_gate_infos),
        )

    @classmethod
    def perfect_star(
        cls,
        num_qubits: int,
        flavour: Type[Flavour],
        comm_instructions: List[Type[NetQASMInstruction]],
        comm_duration: float,
        mem_instructions: List[Type[NetQASMInstruction]],
        mem_duration: float,
        two_instructions: List[Type[NetQASMInstruction]],
        two_duration: float,
        latencies: Optional[EhiLatencies] = None,
    ) -> EhiNodeInfo:
        comm_qubit_info = cls.perfect_qubit(is_communication=True)
        mem_qubit_info = cls.perfect_qubit(is_communication=False)
        comm_gate_infos = cls.perfect_gates(comm_duration, comm_instructions)
        mem_gate_infos = cls.perfect_gates(mem_duration, mem_instructions)
        two_gate_infos = cls.perfect_gates(two_duration, two_instructions)

        q_infos = {0: comm_qubit_info}
        for i in range(1, num_qubits):
            q_infos[i] = mem_qubit_info

        sg_infos = {0: comm_gate_infos}
        for i in range(1, num_qubits):
            sg_infos[i] = mem_gate_infos

        mg_infos = {}
        for i in range(1, num_qubits):
            mg_infos[MultiQubit([0, i])] = two_gate_infos

        if latencies is None:
            latencies = EhiLatencies.all_zero()
        return EhiNodeInfo(q_infos, flavour, sg_infos, mg_infos, latencies)

    @classmethod
    def generic_t1t2_star(
        cls,
        num_qubits: int,
        flavour: Type[Flavour],
        comm_decoherence: float,
        mem_decoherence: float,
        comm_instructions: List[Type[NetQASMInstruction]],
        comm_duration: float,
        comm_instr_decoherence: float,
        mem_instructions: List[Type[NetQASMInstruction]],
        mem_duration: float,
        mem_instr_decoherence: float,
        two_instructions: List[Type[NetQASMInstruction]],
        two_duration: float,
        two_instr_decoherence: float,
        latencies: Optional[EhiLatencies] = None,
    ) -> EhiNodeInfo:
        comm_qubit_info = cls.decoherence_qubit(
            is_communication=True, decoherence_rate=comm_decoherence
        )
        mem_qubit_info = cls.decoherence_qubit(
            is_communication=False, decoherence_rate=mem_decoherence
        )
        comm_gate_infos = cls.decoherence_gates(
            comm_duration, comm_instructions, comm_instr_decoherence
        )
        mem_gate_infos = cls.decoherence_gates(
            mem_duration, mem_instructions, mem_instr_decoherence
        )
        two_gate_infos = cls.decoherence_gates(
            two_duration, two_instructions, two_instr_decoherence
        )

        q_infos = {0: comm_qubit_info}
        for i in range(1, num_qubits):
            q_infos[i] = mem_qubit_info

        sg_infos = {0: comm_gate_infos}
        for i in range(1, num_qubits):
            sg_infos[i] = mem_gate_infos

        mg_infos = {}
        for i in range(1, num_qubits):
            mg_infos[MultiQubit([0, i])] = two_gate_infos

        if latencies is None:
            latencies = EhiLatencies.all_zero()
        return EhiNodeInfo(q_infos, flavour, sg_infos, mg_infos, latencies)


@dataclass(frozen=True)
class UnitModule:
    """Description of virtual memory space for programs. Target for a compiler.

    Simply wraps around a EhiNodeInfo object and provides convenience methods.

    Unit Modules should be used as the interface for compilers and schedulers,
    as well as the program itself. This object does not contain information about
    runtime values (i.e. qubit mappings); this is managed by the Memory Manager.
    Only the Memory Manager should use an EhiNodeInfo object itself,
    namely the object that represents the full quantum memory space of the node."""

    info: EhiNodeInfo

    def is_communication(self, qubit_id: int) -> bool:
        return self.info.qubit_infos[qubit_id].is_communication

    @classmethod
    def from_ehi(cls, ehi: EhiNodeInfo, qubit_ids: List[int]) -> UnitModule:
        """Get a subset of an EhiNodeInfo"""
        for id in qubit_ids:
            if id not in ehi.qubit_infos:
                raise ValueError(f"Qubit ID {id} not in EhiNodeInfo")

        qubit_infos = {i: ehi.qubit_infos[i] for i in qubit_ids}
        single_gate_infos = {i: ehi.single_gate_infos[i] for i in qubit_ids}
        multi_gate_infos = {
            ids: info
            for (ids, info) in ehi.multi_gate_infos.items()
            if all(id in qubit_ids for id in ids.qubit_ids)
        }
        all_qubit_gate_infos: Optional[List[EhiGateInfo]] = None
        if len(ehi.qubit_infos.keys()) == len(qubit_ids):
            all_qubit_gate_infos = ehi.all_qubit_gate_infos

        return UnitModule(
            info=EhiNodeInfo(
                qubit_infos,
                ehi.flavour,
                single_gate_infos,
                multi_gate_infos,
                ehi.latencies,
                all_qubit_gate_infos=all_qubit_gate_infos,
            )
        )

    @classmethod
    def from_full_ehi(cls, ehi: EhiNodeInfo) -> UnitModule:
        """Use the full EhiNodeInfo"""
        return UnitModule(info=copy.deepcopy(ehi))

    def get_all_qubit_ids(self) -> List[int]:
        return list(self.info.qubit_infos.keys())


@dataclass
class EhiNetworkTimebin:
    nodes: FrozenSet[int]
    pids: Dict[int, int]  # node ID -> PID


@dataclass
class EhiNetworkSchedule:
    bin_length: int
    first_bin: int

    bin_pattern: List[EhiNetworkTimebin]
    repeat_period: int

    def next_bin(self, time: int) -> Tuple[int, EhiNetworkTimebin]:
        global_offset = time - self.first_bin

        # Get the start of the current iteration of the repeating pattern.
        curr_pattern_index = floor(global_offset / self.repeat_period)
        curr_pattern_start = curr_pattern_index * self.repeat_period + self.first_bin

        # Get relative time within the pattern.
        time_since_pattern_start = time - curr_pattern_start

        # Get the index of the next bin within the pattern.
        # It could be that we're already in the last bin. Then the next bin
        # is the first bin of the next pattern repetition.
        next_bin_index = ceil(time_since_pattern_start / self.bin_length)

        if next_bin_index >= len(self.bin_pattern):
            next_bin_start = curr_pattern_start + self.repeat_period
            next_bin = self.bin_pattern[0]
        else:
            next_bin_start = next_bin_index * self.bin_length + curr_pattern_start
            next_bin = self.bin_pattern[next_bin_index]
        return next_bin_start, next_bin

    def next_specific_bin(self, time: int, bin: EhiNetworkTimebin) -> int:
        bin_index: Optional[int] = None

        for i, pat_bin in enumerate(self.bin_pattern):
            if bin == pat_bin:
                bin_index = i
        if bin_index is None:
            raise ValueError

        bin_rel_to_pat_start = bin_index * self.bin_length

        # TODO: merge below code with that in the `next_bin()` method
        global_offset = time - self.first_bin
        curr_pattern_index = floor(global_offset / self.repeat_period)
        curr_pattern_start = curr_pattern_index * self.repeat_period + self.first_bin
        time_since_pattern_start = time - curr_pattern_start
        if bin_rel_to_pat_start >= time_since_pattern_start:
            return bin_rel_to_pat_start - time_since_pattern_start
        else:
            return self.repeat_period - time_since_pattern_start + bin_rel_to_pat_start


@dataclass
class EhiNetworkInfo:
    nodes: Dict[int, str]  # node ID -> node name

    # (node A ID, node B ID) -> link info
    # for a pair (a, b) there exists no separate (b, a) info (it is the same)
    links: Dict[FrozenSet[int], EhiLinkInfo]

    network_schedule: Optional[EhiNetworkSchedule] = None

    @classmethod
    def only_nodes(cls, nodes: Dict[int, str]) -> EhiNetworkInfo:
        return EhiNetworkInfo(nodes, {})

    @classmethod
    def fully_connected(
        cls, nodes: Dict[int, str], info: EhiLinkInfo
    ) -> EhiNetworkInfo:
        links: Dict[FrozenSet[int], EhiLinkInfo] = {}
        for n1, n2 in itertools.combinations(nodes.keys(), 2):
            node_link = frozenset([n1, n2])
            links[node_link] = info
        return EhiNetworkInfo(nodes, links)

    @classmethod
    def perfect_fully_connected(
        cls, nodes: Dict[int, str], duration: float
    ) -> EhiNetworkInfo:
        link = EhiLinkInfo(duration=duration, fidelity=1.0)
        return cls.fully_connected(nodes, link)

    def get_node_id(self, name: str) -> int:
        for id, node_name in self.nodes.items():
            if node_name == name:
                return id
        raise ValueError(f"Node with name {name} not found")

    def add_link(self, node1_id: int, node2_id: int, link_info: EhiLinkInfo) -> None:
        if node1_id not in self.nodes:
            raise ValueError(f"Node with ID {node1_id} not found")
        if node2_id not in self.nodes:
            raise ValueError(f"Node with ID {node2_id} not found")
        if node1_id == node2_id:
            raise ValueError("Cannot add link between same node")
        node_link = frozenset([node1_id, node2_id])
        if node_link in self.links:
            raise ValueError(
                f"Link between nodes {node1_id} and {node2_id} already exists"
            )
        self.links[node_link] = link_info

    def get_all_node_names(self) -> List[str]:
        return list(self.nodes.values())

    def get_link(self, node_id1: int, node_id2: int) -> EhiLinkInfo:
        node_link = frozenset([node_id1, node_id2])
        try:
            return self.links[node_link]
        except KeyError:
            raise ValueError(f"No link between nodes {node_id1} and {node_id2}")



from __future__ import annotations

import os
from dataclasses import dataclass

import netsquid as ns

from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    LatenciesConfig,
    NtfConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfigBuilder,
)
from qoala.runtime.program import ProgramInput
from qoala.util.runner import run_single_node_app


def get_config(t1: int = 0, t2: int = 0, gate_fidelity: float = 1.0) -> ProcNodeConfig:
    topology = (
        TopologyConfigBuilder()
        .uniform_topology()
        .num_qubits(1)
        .qubit_t1(t1)
        .qubit_t2(t2)
        .default_generic_gates()
        .zero_gate_durations()
        .perfect_gate_fidelities()
        .all_comm_gates_fidelity(gate_fidelity)
        .build()
    )
    return ProcNodeConfig(
        node_name="alice",
        node_id=0,
        topology=topology,
        latencies=LatenciesConfig(qnos_instr_time=1000),
        ntf=NtfConfig.from_cls_name("GenericNtf"),
    )


def load_program(name: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), name)
    with open(path) as file:
        text = file.read()
    program = QoalaParser(text).parse()

    return program


@dataclass
class BusyProgramTomography:
    prob_x_0: float
    prob_y_0: float
    prob_z_0: float


def run_busy_program(num_iterations: int, node_cfg: ProcNodeConfig):
    network_cfg = ProcNodeNetworkConfig(nodes=[node_cfg], links=[])
    program = load_program("busy_program.iqoala")

    app_results = run_single_node_app(
        num_iterations=num_iterations,
        program_name="alice",
        program=program,
        program_input=ProgramInput.empty(),
        network_cfg=network_cfg,
    )

    all_results = app_results.batch_results["alice"].results
    x_outcomes = [result.values["m_x"] for result in all_results]
    y_outcomes = [result.values["m_y"] for result in all_results]
    z_outcomes = [result.values["m_z"] for result in all_results]

    prob_x_0 = sum(1 - x for x in x_outcomes) / num_iterations
    prob_y_0 = sum(1 - y for y in y_outcomes) / num_iterations
    prob_z_0 = sum(1 - z for z in z_outcomes) / num_iterations
    return BusyProgramTomography(prob_x_0, prob_y_0, prob_z_0)


def test_busy():
    ns.sim_reset()
    # LogManager.set_log_level("DEBUG")

    node_cfg = get_config()
    result = run_busy_program(200, node_cfg)

    print(f"prob_x_0: {result.prob_x_0}")
    print(f"prob_y_0: {result.prob_y_0}")
    print(f"prob_z_0: {result.prob_z_0}")

    assert 0.8 < result.prob_x_0
    assert 0.3 < result.prob_y_0 < 0.7
    assert 0.3 < result.prob_z_0 < 0.7


def test_busy_bad_coherence():
    ns.sim_reset()
    # LogManager.set_log_level("DEBUG")

    node_cfg = get_config(t1=10, t2=10)
    result = run_busy_program(200, node_cfg)

    print(f"prob_x_0: {result.prob_x_0}")
    print(f"prob_y_0: {result.prob_y_0}")
    print(f"prob_z_0: {result.prob_z_0}")
    # Because of decoherence, we expect the prepared state to be close to |0>
    # (i.e. Z-outcome should be mostly 0, X and Y uniformly random)
    assert 0.3 < result.prob_x_0 < 0.7
    assert 0.3 < result.prob_y_0 < 0.7
    assert 0.8 < result.prob_z_0


def test_busy_bad_gates():
    ns.sim_reset()
    # LogManager.set_log_level("DEBUG")

    node_cfg = get_config(gate_fidelity=0.5)
    result = run_busy_program(200, node_cfg)

    print(f"prob_x_0: {result.prob_x_0}")
    print(f"prob_y_0: {result.prob_y_0}")
    print(f"prob_z_0: {result.prob_z_0}")
    # Because of gate noise, the prepared state should be close to maximally mixed
    # (i.e. all outcomes should be uniformly random)
    assert 0.3 < result.prob_x_0 < 0.7
    assert 0.3 < result.prob_y_0 < 0.7
    assert 0.3 < result.prob_z_0 < 0.7


if __name__ == "__main__":
    test_busy()
    test_busy_bad_coherence()
    test_busy_bad_gates()



from __future__ import annotations

import datetime
import json
import os
import time
from argparse import ArgumentParser
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import List

import netsquid as ns
from netqasm.lang.instr.flavour import NVFlavour, TrappedIonFlavour

from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    LatenciesConfig,
    NtfConfig,
    NvParams,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.program import BatchResult, ProgramInput
from qoala.util.runner import run_two_node_app


def relative_to_cwd(file: str) -> str:
    return os.path.join(os.path.dirname(__file__), file)


def create_procnode_cfg(
    name: str, id: int, num_qubits: int, hardware: str
) -> ProcNodeConfig:
    if hardware == "generic":
        return ProcNodeConfig(
            node_name=name,
            node_id=id,
            topology=TopologyConfig.perfect_config_uniform_default_params(num_qubits),
            latencies=LatenciesConfig(qnos_instr_time=1000),
            ntf=NtfConfig.from_cls_name("GenericNtf"),
        )
    elif hardware == "nv":
        return ProcNodeConfig(
            node_name=name,
            node_id=id,
            topology=TopologyConfig.from_nv_params(
                num_qubits=num_qubits, params=NvParams()
            ),
            latencies=LatenciesConfig(qnos_instr_time=1000),
            ntf=NtfConfig.from_cls_name("NvNtf"),
        )
    elif hardware == "tri":
        return ProcNodeConfig(
            node_name=name,
            node_id=id,
            topology=TopologyConfig.perfect_tri_default_params(num_qubits=num_qubits),
            latencies=LatenciesConfig(qnos_instr_time=1000),
            ntf=NtfConfig.from_cls_name("TrappedIonNtf"),
        )


def load_program(path: str, hardware: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()

    if hardware == "generic":
        flavour = None
    elif hardware == "nv":
        flavour = NVFlavour()
    elif hardware == "tri":
        flavour = TrappedIonFlavour()
    return QoalaParser(text, flavour=flavour).parse()


@dataclass
class QkdResult:
    alice_result: BatchResult
    bob_result: BatchResult


def run_qkd(
    hardware: str,
    num_iterations: int,
    num_pairs: int,
    linear: bool = True,
):
    num_qubits = 3
    alice_id = 0
    bob_id = 1

    alice_node_cfg = create_procnode_cfg("alice", alice_id, num_qubits, hardware)
    bob_node_cfg = create_procnode_cfg("bob", bob_id, num_qubits, hardware)

    network_cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
        nodes=[alice_node_cfg, bob_node_cfg], link_duration=1000
    )

    alice_file = "qkd_npairs_MD_alice.iqoala"
    bob_file = "qkd_npairs_MD_bob.iqoala"
    alice_program = load_program(alice_file, hardware)
    bob_program = load_program(bob_file, hardware)

    alice_input = ProgramInput({"bob_id": bob_id, "N": num_pairs})
    bob_input = ProgramInput({"alice_id": alice_id, "N": num_pairs})

    app_result = run_two_node_app(
        num_iterations=num_iterations,
        programs={"alice": alice_program, "bob": bob_program},
        program_inputs={"alice": alice_input, "bob": bob_input},
        network_cfg=network_cfg,
        linear=linear,
    )

    alice_result = app_result.batch_results["alice"]
    bob_result = app_result.batch_results["bob"]

    return QkdResult(alice_result, bob_result)


def qkd_npairs_md(hardware: str, num_iterations: int, num_pairs: int):
    num_pairs = int(num_pairs)
    ns.sim_reset()

    qkd_result = run_qkd(hardware, num_iterations, num_pairs=num_pairs)

    alice_results = qkd_result.alice_result.results
    bob_results = qkd_result.bob_result.results

    assert len(alice_results) == num_iterations
    assert len(bob_results) == num_iterations

    alice_outcomes = [alice_results[i].values for i in range(num_iterations)]
    bob_outcomes = [bob_results[i].values for i in range(num_iterations)]

    for alice, bob in zip(alice_outcomes, bob_outcomes):
        # print(f"alice: {alice['outcomes']}")
        # print(f"bob  : {bob['outcomes']}")
        assert alice["outcomes"] == bob["outcomes"]


@dataclass
class DataPoint:
    success: bool
    sim_duration: float


@dataclass
class DataMeta:
    timestamp: str
    num_iterations: int
    num_pairs: int


@dataclass
class Data:
    meta: DataMeta
    data_points: List[DataPoint]


def run_hardware(hardware: str, num_iterations: int, num_pairs: int) -> DataPoint:
    start = time.time()

    success: bool
    try:
        qkd_npairs_md(hardware, num_iterations, num_pairs)
        success = True
    except AssertionError:
        success = False

    end = time.time()
    duration = round(end - start, 2)
    print(f"duration: {duration} s")

    return DataPoint(success, duration)


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--num_iterations", "-n", type=int, required=True)
    parser.add_argument("--num_pairs", "-p", type=int, required=True)

    args = parser.parse_args()
    num_iterations = args.num_iterations
    num_pairs = args.num_pairs

    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")

    data_points: List[DataPoint] = []
    for hardware in ["generic", "nv", "tri"]:
        data_point = run_hardware(hardware, num_iterations, num_pairs)
        data_points.append(data_point)

    abs_dir = relative_to_cwd(f"data")
    Path(abs_dir).mkdir(parents=True, exist_ok=True)
    last_path = os.path.join(abs_dir, "LAST.json")
    timestamp_path = os.path.join(abs_dir, f"{timestamp}.json")

    meta = DataMeta(
        timestamp=timestamp, num_iterations=num_iterations, num_pairs=num_pairs
    )
    data = Data(meta=meta, data_points=data_points)
    json_data = asdict(data)

    with open(last_path, "w") as datafile:
        json.dump(json_data, datafile)
    with open(timestamp_path, "w") as datafile:
        json.dump(json_data, datafile)



from netqasm.lang.instr.core import MeasInstruction, RetRegInstruction, SetInstruction
from netqasm.lang.instr.vanilla import RotZInstruction
from netqasm.lang.operand import Register, Template
from netqasm.lang.subroutine import Subroutine

from qoala.lang.hostlang import (
    AddCValueOp,
    AssignCValueOp,
    BasicBlock,
    BasicBlockType,
    IqoalaSingleton,
    IqoalaTuple,
    IqoalaVector,
    ReceiveCMsgOp,
    ReturnResultOp,
    RunSubroutineOp,
    SendCMsgOp,
)
from qoala.lang.program import LocalRoutine, ProgramMeta, QoalaProgram
from qoala.lang.routine import RoutineMetadata
from qoala.util.tests import text_equal


def test_serialize_meta_1():
    expected = """
META_START
name: alice
parameters: 
csockets: 0 -> bob
epr_sockets: 
META_END
    """

    meta = ProgramMeta(name="alice", parameters=[], csockets={0: "bob"}, epr_sockets={})
    assert text_equal(meta.serialize(), expected)


def test_serialize_meta_2():
    expected = """
META_START
name: alice
parameters: theta1, theta2
csockets: 0 -> bob, 1 -> charlie
epr_sockets: 1 -> charlie
META_END
    """

    meta = ProgramMeta(
        name="alice",
        parameters=["theta1", "theta2"],
        csockets={0: "bob", 1: "charlie"},
        epr_sockets={1: "charlie"},
    )
    assert text_equal(meta.serialize(), expected)


def test_serialize_host_code_1():
    expected = """
^b0 {type = CL}:
    my_value = assign_cval() : 1
    remote_id = assign_cval() : 0
    send_cmsg(remote_id, my_value)
    received_value = recv_cmsg(remote_id)
    new_value = assign_cval() : 3
    my_value = add_cval_c(new_value, new_value)
^b1 {type = QL}:
    tuple<m> = run_subroutine(tuple<my_value>) : subrt1
    x<5> = run_subroutine() : subrt2
^b2 {type = CL}:
    return_result(m)
    """

    b0 = BasicBlock(
        "b0",
        BasicBlockType.CL,
        instructions=[
            AssignCValueOp(IqoalaSingleton("my_value"), 1),
            AssignCValueOp(IqoalaSingleton("remote_id"), 0),
            SendCMsgOp(IqoalaSingleton("remote_id"), IqoalaSingleton("my_value")),
            ReceiveCMsgOp(
                IqoalaSingleton("remote_id"), IqoalaSingleton("received_value")
            ),
            AssignCValueOp(IqoalaSingleton("new_value"), 3),
            AddCValueOp(
                IqoalaSingleton("my_value"),
                IqoalaSingleton("new_value"),
                IqoalaSingleton("new_value"),
            ),
        ],
    )
    b1 = BasicBlock(
        "b1",
        BasicBlockType.QL,
        instructions=[
            RunSubroutineOp(IqoalaTuple(["m"]), IqoalaTuple(["my_value"]), "subrt1"),
            RunSubroutineOp(IqoalaVector("x", 5), IqoalaTuple([]), "subrt2"),
        ],
    )
    b2 = BasicBlock(
        "b2",
        BasicBlockType.CL,
        instructions=[
            ReturnResultOp(IqoalaSingleton("m")),
        ],
    )

    program = QoalaProgram(meta=ProgramMeta.empty("alice"), blocks=[b0, b1, b2])
    print(program.serialize_host_code())
    assert text_equal(program.serialize_host_code(), expected)


def test_serialize_subroutines_1():
    expected = """
SUBROUTINE subrt1
    params: my_value
    returns: m
    uses: 0
    keeps: 
  NETQASM_START
    set Q0 0
    rot_z Q0 {my_value} 4
    meas Q0 M0
    ret_reg M0
  NETQASM_END
    """

    Q0 = Register.from_str("Q0")
    M0 = Register.from_str("M0")
    subrt = LocalRoutine(
        name="subrt1",
        subroutine=Subroutine(
            instructions=[
                SetInstruction(reg=Q0, imm=0),
                RotZInstruction(reg=Q0, imm0=Template("my_value"), imm1=4),
                MeasInstruction(reg0=Q0, reg1=M0),
                RetRegInstruction(reg=M0),
            ],
            arguments=["my_value"],
        ),
        return_vars=["m"],
        metadata=RoutineMetadata.free_all([0]),
    )
    program = QoalaProgram(
        meta=ProgramMeta.empty("alice"),
        blocks=[],
        local_routines={"subrt1": subrt},
    )
    assert text_equal(program.serialize_subroutines(), expected)


def test_serialize_subroutines_2():
    expected = """
SUBROUTINE subrt1
    params: param1
    returns: outcomes<10>
    uses: 0
    keeps: 
  NETQASM_START
    set R0 {param1}
    meas Q0 M0
  NETQASM_END

SUBROUTINE subrt2
    params: theta
    returns: 
    uses:
    keeps: 
  NETQASM_START
    set R0 {theta}
  NETQASM_END
    """

    R0 = Register.from_str("R0")
    Q0 = Register.from_str("Q0")
    M0 = Register.from_str("M0")
    subrt1 = LocalRoutine(
        name="subrt1",
        subroutine=Subroutine(
            instructions=[
                SetInstruction(reg=R0, imm=Template("param1")),
                MeasInstruction(reg0=Q0, reg1=M0),
            ],
            arguments=["param1"],
        ),
        return_vars=[IqoalaVector("outcomes", 10)],
        metadata=RoutineMetadata.free_all([0]),
    )
    subrt2 = LocalRoutine(
        name="subrt2",
        subroutine=Subroutine(
            instructions=[
                SetInstruction(reg=R0, imm=Template("theta")),
            ],
            arguments=["theta"],
        ),
        return_vars=[],
        metadata=RoutineMetadata.use_none(),
    )
    program = QoalaProgram(
        meta=ProgramMeta.empty("alice"),
        blocks=[],
        local_routines={"subrt1": subrt1, "subrt2": subrt2},
    )
    assert text_equal(program.serialize_subroutines(), expected)


def test_serialize_program():
    expected = """
META_START
name: alice
parameters: 
csockets: 0 -> bob
epr_sockets: 
META_END

^b0 {type = CL}:
    my_value = assign_cval() : 1
    remote_id = assign_cval() : 0
    send_cmsg(remote_id, my_value)
    received_value = recv_cmsg(remote_id)
    new_value = assign_cval() : 3
    my_value = add_cval_c(new_value, new_value)
^b1 {type = QL}:
    tuple<m> = run_subroutine(tuple<my_value>) : subrt1
^b2 {type = CL}:
    return_result(m)

SUBROUTINE subrt1
    params: my_value
    returns: m
    uses: 0
    keeps: 
  NETQASM_START
    set Q0 0
    rot_z Q0 {my_value} 4
    meas Q0 M0
    ret_reg M0
  NETQASM_END
    """

    meta = ProgramMeta(name="alice", parameters=[], csockets={0: "bob"}, epr_sockets={})
    b0_instructions = [
        AssignCValueOp("my_value", 1),
        AssignCValueOp("remote_id", 0),
        SendCMsgOp("remote_id", "my_value"),
        ReceiveCMsgOp("remote_id", "received_value"),
        AssignCValueOp("new_value", 3),
        AddCValueOp("my_value", "new_value", "new_value"),
    ]
    b1_instructions = [
        RunSubroutineOp(IqoalaTuple(["m"]), IqoalaTuple(["my_value"]), "subrt1")
    ]
    b2_instructions = [ReturnResultOp("m")]
    b0 = BasicBlock("b0", BasicBlockType.CL, b0_instructions)
    b1 = BasicBlock("b1", BasicBlockType.QL, b1_instructions)
    b2 = BasicBlock("b2", BasicBlockType.CL, b2_instructions)
    Q0 = Register.from_str("Q0")
    M0 = Register.from_str("M0")
    subrt = LocalRoutine(
        name="subrt1",
        subroutine=Subroutine(
            instructions=[
                SetInstruction(reg=Q0, imm=0),
                RotZInstruction(reg=Q0, imm0=Template("my_value"), imm1=4),
                MeasInstruction(reg0=Q0, reg1=M0),
                RetRegInstruction(reg=M0),
            ],
            arguments=["my_value"],
        ),
        return_vars=["m"],
        metadata=RoutineMetadata.free_all([0]),
    )

    program = QoalaProgram(
        meta=meta, blocks=[b0, b1, b2], local_routines={"subrt1": subrt}
    )

    assert text_equal(program.serialize(), expected)


if __name__ == "__main__":
    test_serialize_meta_1()
    test_serialize_meta_2()
    test_serialize_host_code_1()
    test_serialize_subroutines_1()
    test_serialize_subroutines_2()
    test_serialize_program()



import pytest

from qoala.runtime.sharedmem import (
    MemAddr,
    SharedMemIllegalRegionError,
    SharedMemNotAllocatedError,
    SharedMemory,
    SharedMemReadError,
)


def test1():
    mgr = SharedMemory()

    params = [0, 1]

    # Not allocated yet
    with pytest.raises(SharedMemNotAllocatedError):
        mgr.write_lr_in(MemAddr(0), params)

    addr = mgr.allocate_lr_in(2)
    mgr.write_lr_in(addr, params)
    assert len(mgr._arrays._memory) == 1
    assert mgr._arrays._memory[addr] == params

    assert mgr._lr_in_addrs == [addr]

    assert mgr.read_lr_in(addr, 2) == params

    assert mgr.read_lr_in(addr, 1) == params[0:1]
    assert mgr.read_lr_in(addr, 1, offset=1) == params[1:]

    # Beyond end of array
    with pytest.raises(SharedMemReadError):
        assert mgr.read_lr_in(addr, 1, offset=2)

    # Too long
    with pytest.raises(SharedMemReadError):
        assert mgr.read_lr_in(addr, 3, offset=0)

    # Wrong type
    with pytest.raises(SharedMemIllegalRegionError):
        mgr.write_cr_in(addr, params)

    addr2 = mgr.allocate_rr_in(3)
    assert len(mgr._arrays._memory) == 2
    params2 = [0, 1, 2]

    with pytest.raises(SharedMemIllegalRegionError):
        assert mgr.write_lr_in(addr2, params2)

    mgr.write_rr_in(addr2, params2)

    assert mgr._arrays._memory[addr2] == params2

    assert mgr._rr_in_addrs == [addr2]

    assert mgr._lr_in_addrs == [addr]
    assert mgr._addr_counter == 2

    assert mgr.read_rr_in(addr2, 3) == params2

    assert mgr.read_rr_in(addr2, 2) == params2[0:2]
    assert mgr.read_rr_in(addr2, 2, offset=1) == params2[1:]

    with pytest.raises(SharedMemIllegalRegionError):
        assert mgr.read_lr_in(addr2, 3, offset=0)


if __name__ == "__main__":
    test1()



from __future__ import annotations

from typing import Generator

import netsquid as ns
from netsquid.components.cchannel import ClassicalChannel
from netsquid.nodes import Node

from pydynaa import EventExpression
from qoala.lang.ehi import EhiNetworkInfo
from qoala.runtime.message import Message
from qoala.sim.host.hostcomp import HostComponent
from qoala.sim.host.hostinterface import HostInterface


def create_hostcomp(num_other_nodes: int) -> HostComponent:
    node = Node(name="alice", ID=0)

    nodes = {id: f"node_{id}" for id in range(1, num_other_nodes + 1)}
    nodes[0] = "alice"
    ehi_network = EhiNetworkInfo.only_nodes(nodes)

    return HostComponent(node, ehi_network)


def test_no_other_nodes():
    comp = create_hostcomp(num_other_nodes=0)

    # should have qnos_{in|out} and ntsk_{in|out} ports
    assert len(comp.ports) == 4
    assert "qnos_in" in comp.ports
    assert "qnos_out" in comp.ports
    assert "nstk_in" in comp.ports
    assert "nstk_out" in comp.ports


def test_one_other_node():
    comp = create_hostcomp(num_other_nodes=1)

    # should have 2 qnos ports + 2 nstk ports + 2 peer ports
    assert len(comp.ports) == 6
    assert "qnos_in" in comp.ports
    assert "qnos_out" in comp.ports
    assert "nstk_in" in comp.ports
    assert "nstk_out" in comp.ports

    assert "peer_node_1_in" in comp.ports
    assert "peer_node_1_out" in comp.ports

    # Test properties
    assert comp.peer_in_port("node_1") == comp.ports["peer_node_1_in"]
    assert comp.peer_out_port("node_1") == comp.ports["peer_node_1_out"]


def test_many_other_nodes():
    comp = create_hostcomp(num_other_nodes=5)

    # should have 2 qnos ports + 2 nstk ports + 5 * 2 peer ports
    assert len(comp.ports) == 14
    assert "qnos_in" in comp.ports
    assert "qnos_out" in comp.ports
    assert "nstk_in" in comp.ports
    assert "nstk_out" in comp.ports

    for i in range(1, 6):
        assert f"peer_node_{i}_in" in comp.ports
        assert f"peer_node_{i}_out" in comp.ports
        # Test properties
        assert comp.peer_in_port(f"node_{i}") == comp.ports[f"peer_node_{i}_in"]
        assert comp.peer_out_port(f"node_{i}") == comp.ports[f"peer_node_{i}_out"]


def test_connection():
    ns.sim_reset()

    alice = Node(name="alice", ID=0)
    bob = Node(name="bob", ID=1)
    ehi_network = EhiNetworkInfo.only_nodes({alice.ID: alice.name, bob.ID: bob.name})

    alice_comp = HostComponent(alice, ehi_network)
    bob_comp = HostComponent(bob, ehi_network)

    alice_comp.peer_out_port("bob").connect(bob_comp.peer_in_port("alice"))
    alice_comp.peer_in_port("bob").connect(bob_comp.peer_out_port("alice"))

    class AliceHostInterface(HostInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            self.send_peer_msg("bob", Message(0, 0, "hello"))

    class BobHostInterface(HostInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            msg = yield from self.receive_peer_msg("alice")
            assert msg.content == "hello"

    alice_intf = AliceHostInterface(alice_comp, ehi_network)
    bob_intf = BobHostInterface(bob_comp, ehi_network)

    alice_intf.start()
    bob_intf.start()

    ns.sim_run()


def test_three_way_connection():
    ns.sim_reset()

    alice = Node(name="alice", ID=0)
    bob = Node(name="bob", ID=1)
    charlie = Node(name="charlie", ID=2)
    ehi_network = EhiNetworkInfo.only_nodes(
        {alice.ID: alice.name, bob.ID: bob.name, charlie.ID: charlie.name}
    )

    alice_comp = HostComponent(alice, ehi_network)
    bob_comp = HostComponent(bob, ehi_network)
    charlie_comp = HostComponent(charlie, ehi_network)

    alice_comp.peer_out_port("bob").connect(bob_comp.peer_in_port("alice"))
    alice_comp.peer_in_port("bob").connect(bob_comp.peer_out_port("alice"))
    alice_comp.peer_out_port("charlie").connect(charlie_comp.peer_in_port("alice"))
    alice_comp.peer_in_port("charlie").connect(charlie_comp.peer_out_port("alice"))
    bob_comp.peer_out_port("charlie").connect(charlie_comp.peer_in_port("bob"))
    bob_comp.peer_in_port("charlie").connect(charlie_comp.peer_out_port("bob"))

    class AliceHostInterface(HostInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            self.send_peer_msg("bob", Message(0, 0, "hello bob"))
            self.send_peer_msg("charlie", Message(0, 0, "hello charlie"))

    class BobHostInterface(HostInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            msg = yield from self.receive_peer_msg("alice")
            assert msg.content == "hello bob"

    class CharlieHostInterface(HostInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            msg = yield from self.receive_peer_msg("alice")
            assert msg.content == "hello charlie"

    alice_intf = AliceHostInterface(alice_comp, ehi_network)
    bob_intf = BobHostInterface(bob_comp, ehi_network)
    charlie_intf = CharlieHostInterface(charlie_comp, ehi_network)

    alice_intf.start()
    bob_intf.start()
    charlie_intf.start()

    ns.sim_run()


def test_connection_with_channel():
    ns.sim_reset()

    alice = Node(name="alice", ID=0)
    bob = Node(name="bob", ID=1)
    ehi_network = EhiNetworkInfo.only_nodes({alice.ID: alice.name, bob.ID: bob.name})

    alice_comp = HostComponent(alice, ehi_network)
    bob_comp = HostComponent(bob, ehi_network)

    channel_ab = ClassicalChannel("chan_ab", delay=1000)
    channel_ba = ClassicalChannel("chan_ba", delay=1000)

    alice_comp.peer_out_port("bob").connect(channel_ab.ports["send"])
    channel_ab.ports["recv"].connect(bob_comp.peer_in_port("alice"))

    bob_comp.peer_out_port("alice").connect(channel_ba.ports["send"])
    channel_ba.ports["recv"].connect(alice_comp.peer_in_port("bob"))

    class AliceHostInterface(HostInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            self.send_peer_msg("bob", Message(0, 0, "hello"))

    class BobHostInterface(HostInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            assert ns.sim_time() == 0
            msg = yield from self.receive_peer_msg("alice")
            assert msg == Message(0, 0, "hello")
            assert ns.sim_time() == 1000

    alice_intf = AliceHostInterface(alice_comp, ehi_network)
    bob_intf = BobHostInterface(bob_comp, ehi_network)

    alice_intf.start()
    bob_intf.start()

    ns.sim_run()


def test_connection_with_pids():
    ns.sim_reset()

    alice = Node(name="alice", ID=0)
    bob = Node(name="bob", ID=1)
    ehi_network = EhiNetworkInfo.only_nodes({alice.ID: alice.name, bob.ID: bob.name})

    alice_comp = HostComponent(alice, ehi_network)
    bob_comp = HostComponent(bob, ehi_network)

    channel_ab = ClassicalChannel("chan_ab", delay=1000)
    channel_ba = ClassicalChannel("chan_ba", delay=1000)

    alice_comp.peer_out_port("bob").connect(channel_ab.ports["send"])
    channel_ab.ports["recv"].connect(bob_comp.peer_in_port("alice"))

    bob_comp.peer_out_port("alice").connect(channel_ba.ports["send"])
    channel_ba.ports["recv"].connect(alice_comp.peer_in_port("bob"))

    class AliceHostInterface(HostInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            # Send message from PID 0 (alice) to PID 0 (bob)
            self.send_peer_msg("bob", Message(0, 0, "hello (0, 0)"))

            yield from self.wait(500)

            # Send message from PID 2 (alice) to PID 1 (bob)
            self.send_peer_msg("bob", Message(2, 1, "hello (2, 1)"))

    class BobHostInterface(HostInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            assert len(self.get_available_messages("alice")) == 0
            assert ns.sim_time() == 0
            yield from self.wait_for_msg("alice")
            assert ns.sim_time() == 1000
            assert self.get_available_messages("alice") == [(0, 0)]
            msg00 = self.pop_msg("alice", 0, 0)
            assert msg00 == Message(0, 0, "hello (0, 0)")

            yield from self.wait_for_msg("alice")
            assert ns.sim_time() == 1500  # 1000 + 500
            assert self.get_available_messages("alice") == [(2, 1)]
            msg21 = self.pop_msg("alice", 2, 1)
            print(f"{self.name}: received msg (0, 0) with content: {msg00.content}")
            print(f"{self.name}: received msg (2, 1) with content: {msg21.content}")
            assert msg21 == Message(2, 1, "hello (2, 1)")

    alice_intf = AliceHostInterface(alice_comp, ehi_network)
    bob_intf = BobHostInterface(bob_comp, ehi_network)

    alice_intf.start()
    bob_intf.start()

    ns.sim_run()


if __name__ == "__main__":
    test_no_other_nodes()
    test_one_other_node()
    test_many_other_nodes()

    test_connection()
    test_three_way_connection()
    test_connection_with_channel()
    test_connection_with_pids()



from __future__ import annotations

import datetime
import json
import os
import random
import time
from argparse import ArgumentParser
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Dict, List

import netsquid as ns
from netqasm.lang.instr.flavour import NVFlavour, TrappedIonFlavour

from qoala.lang.ehi import UnitModule
from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    ClassicalConnectionConfig,
    LatenciesConfig,
    NtfConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.program import BatchResult, ProgramBatch, ProgramInput
from qoala.runtime.statistics import SchedulerStatistics
from qoala.sim.build import build_network_from_config
from qoala.util.runner import AppResult, create_batch


def relative_to_cwd(file: str) -> str:
    return os.path.join(os.path.dirname(__file__), file)


def create_procnode_cfg(
    name: str, id: int, num_qubits: int, hardware: str
) -> ProcNodeConfig:
    if hardware == "generic":
        return ProcNodeConfig(
            node_name=name,
            node_id=id,
            topology=TopologyConfig.perfect_config_uniform_default_params(num_qubits),
            latencies=LatenciesConfig(qnos_instr_time=1000),
            ntf=NtfConfig.from_cls_name("GenericNtf"),
        )
    elif hardware == "nv":
        return ProcNodeConfig(
            node_name=name,
            node_id=id,
            topology=TopologyConfig.perfect_nv_default_params(num_qubits=num_qubits),
            latencies=LatenciesConfig(qnos_instr_time=1000),
            ntf=NtfConfig.from_cls_name("NvNtf"),
        )
    elif hardware == "tri":
        return ProcNodeConfig(
            node_name=name,
            node_id=id,
            topology=TopologyConfig.perfect_tri_default_params(num_qubits=num_qubits),
            latencies=LatenciesConfig(qnos_instr_time=1000),
            ntf=NtfConfig.from_cls_name("TrappedIonNtf"),
        )


def load_program(path: str, hardware: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()

    if hardware == "generic":
        flavour = None
    elif hardware == "nv":
        flavour = NVFlavour()
    elif hardware == "tri":
        flavour = TrappedIonFlavour()
    return QoalaParser(text, flavour=flavour).parse()


@dataclass
class GhzResult:
    alice_results: BatchResult
    bob_results: BatchResult
    charlie_results: BatchResult


def run_ghz(hardware: str, num_iterations: int) -> GhzResult:
    ns.sim_reset()

    alice_id = 0
    bob_id = 1
    charlie_id = 2

    alice_node_cfg = create_procnode_cfg("alice", alice_id, 1, hardware)
    bob_node_cfg = create_procnode_cfg("bob", bob_id, 2, hardware)
    charlie_node_cfg = create_procnode_cfg("charlie", charlie_id, 2, hardware)

    cconn_ab = ClassicalConnectionConfig.from_nodes(alice_id, bob_id, 1e6)
    cconn_ac = ClassicalConnectionConfig.from_nodes(alice_id, charlie_id, 1e6)
    cconn_bc = ClassicalConnectionConfig.from_nodes(bob_id, charlie_id, 1e6)
    network_cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
        nodes=[alice_node_cfg, bob_node_cfg, charlie_node_cfg], link_duration=1000
    )
    network_cfg.cconns = [cconn_ab, cconn_ac, cconn_bc]

    alice_file = f"ghz_{hardware}_alice.iqoala"
    bob_file = f"ghz_{hardware}_bob.iqoala"
    charlie_file = f"ghz_{hardware}_charlie.iqoala"
    alice_program = load_program(alice_file, hardware)
    bob_program = load_program(bob_file, hardware)
    charlie_program = load_program(charlie_file, hardware)

    alice_input = [
        ProgramInput({"bob_id": bob_id, "charlie_id": charlie_id})
        for _ in range(num_iterations)
    ]
    bob_input = [
        ProgramInput({"alice_id": alice_id, "charlie_id": charlie_id})
        for _ in range(num_iterations)
    ]
    charlie_input = [
        ProgramInput({"alice_id": alice_id, "bob_id": bob_id})
        for _ in range(num_iterations)
    ]

    ns.sim_reset()
    ns.set_qstate_formalism(ns.QFormalism.DM)
    seed = random.randint(0, 1000)
    ns.set_random_state(seed=seed)

    network = build_network_from_config(network_cfg)
    names = ["alice", "bob", "charlie"]
    programs = {"alice": alice_program, "bob": bob_program, "charlie": charlie_program}
    program_inputs = {"alice": alice_input, "bob": bob_input, "charlie": charlie_input}
    batches: Dict[str, ProgramBatch] = {}  # node -> batch

    for name in names:
        procnode = network.nodes[name]
        program = programs[name]
        inputs = program_inputs[name]

        unit_module = UnitModule.from_full_ehi(procnode.memmgr.get_ehi())
        batch_info = create_batch(program, unit_module, inputs, num_iterations)
        batches[name] = procnode.submit_batch(batch_info)

    for batch in batches.values():
        assert batch.batch_id == 0
        assert [p.pid for p in batch.instances] == [i for i in range(num_iterations)]

    for name in names:
        procnode = network.nodes[name]
        remote_pids = {0: [i for i in range(num_iterations)]}
        procnode.initialize_processes(remote_pids=remote_pids, linear=True)

    network.start()
    ns.sim_run()

    results: Dict[str, BatchResult] = {}
    statistics: Dict[str, SchedulerStatistics] = {}

    for name in names:
        procnode = network.nodes[name]
        # only one batch (ID = 0), so get value at index 0
        results[name] = procnode.scheduler.get_batch_results()[0]
        statistics[name] = procnode.scheduler.get_statistics()

    total_duration = ns.sim_time()
    app_result = AppResult(results, statistics, total_duration)
    print(f"{app_result.total_duration:_}")

    alice_result = app_result.batch_results["alice"]
    bob_result = app_result.batch_results["bob"]
    charlie_result = app_result.batch_results["charlie"]

    return GhzResult(alice_result, bob_result, charlie_result)


@dataclass
class DataPoint:
    hardware: str
    success: bool
    sim_duration: float


@dataclass
class DataMeta:
    timestamp: str
    num_iterations: int


@dataclass
class Data:
    meta: DataMeta
    data_points: List[DataPoint]


def test_ghz(hardware: str, num_iterations: int) -> None:
    result = run_ghz(hardware, num_iterations)

    alice_results = result.alice_results.results
    alice_outcomes = [result.values["outcome"] for result in alice_results]
    # print(alice_outcomes)

    bob_results = result.bob_results.results
    bob_outcomes = [result.values["outcome"] for result in bob_results]
    # print(bob_outcomes)

    charlie_results = result.charlie_results.results
    charlie_outcomes = [result.values["outcome"] for result in charlie_results]
    # print(charlie_outcomes)

    assert alice_outcomes == bob_outcomes == charlie_outcomes


def run_hardware(hardware: str, num_iterations: int) -> DataPoint:
    start = time.time()

    success: bool
    try:
        test_ghz(hardware, num_iterations)
        success = True
    except AssertionError:
        success = False

    end = time.time()
    duration = round(end - start, 2)
    print(f"duration: {duration} s")

    return DataPoint(hardware, success, duration)


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--num_iterations", "-n", type=int, required=True)

    args = parser.parse_args()
    num_iterations = args.num_iterations

    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")

    data_points: List[DataPoint] = []
    for hardware in ["generic", "nv", "tri"]:
        data_point = run_hardware(hardware, num_iterations)
        data_points.append(data_point)

    abs_dir = relative_to_cwd(f"data")
    Path(abs_dir).mkdir(parents=True, exist_ok=True)
    last_path = os.path.join(abs_dir, "LAST.json")
    timestamp_path = os.path.join(abs_dir, f"{timestamp}.json")

    meta = DataMeta(timestamp=timestamp, num_iterations=num_iterations)
    data = Data(meta=meta, data_points=data_points)
    json_data = asdict(data)

    with open(last_path, "w") as datafile:
        json.dump(json_data, datafile)
    with open(timestamp_path, "w") as datafile:
        json.dump(json_data, datafile)



from __future__ import annotations

from typing import Dict

from netsquid.components.component import Component, Port

from qoala.lang.ehi import EhiNetworkInfo


class EntDistComponent(Component):
    def __init__(self, ehi_network: EhiNetworkInfo) -> None:
        super().__init__("global_entanglement_distributor")

        self._node_in_ports: Dict[str, str] = {}  # node name -> port name
        self._node_out_ports: Dict[str, str] = {}  # node name -> port name

        for node_name in ehi_network.nodes.values():
            port_in_name = f"node_{node_name}_in"
            port_out_name = f"node_{node_name}_out"
            self._node_in_ports[node_name] = port_in_name
            self._node_out_ports[node_name] = port_out_name

        self.add_ports(self._node_in_ports.values())
        self.add_ports(self._node_out_ports.values())

    def node_in_port(self, name: str) -> Port:
        port_name = self._node_in_ports[name]
        return self.ports[port_name]

    def node_out_port(self, name: str) -> Port:
        port_name = self._node_out_ports[name]
        return self.ports[port_name]



from qoala.sim.netstack.netstack import Netstack
from qoala.sim.netstack.netstackcomp import NetstackComponent
from qoala.sim.netstack.netstackinterface import NetstackInterface, NetstackLatencies
from qoala.sim.netstack.netstackprocessor import NetstackProcessor



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Optional

import netsquid as ns

from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    LatenciesConfig,
    LinkBetweenNodesConfig,
    LinkConfig,
    NtfConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.program import BatchResult, ProgramInput
from qoala.util.runner import run_two_node_app


def create_procnode_cfg(name: str, id: int, num_qubits: int) -> ProcNodeConfig:
    return ProcNodeConfig(
        node_name=name,
        node_id=id,
        topology=TopologyConfig.perfect_config_uniform_default_params(num_qubits),
        latencies=LatenciesConfig(qnos_instr_time=1000),
        ntf=NtfConfig.from_cls_name("GenericNtf"),
    )


def load_program(path: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()
    return QoalaParser(text).parse()


@dataclass
class QkdResult:
    alice_result: BatchResult
    bob_result: BatchResult


def run_qkd(
    num_iterations: int,
    alice_file: str,
    bob_file: str,
    prob_max_mixed: float,
    attempt_success_prob: float,
    attempt_duration: float,
    state_delay: float,
    num_pairs: Optional[int] = None,
):
    num_qubits = 3
    alice_id = 0
    bob_id = 1

    alice_node_cfg = create_procnode_cfg("alice", alice_id, num_qubits)
    bob_node_cfg = create_procnode_cfg("bob", bob_id, num_qubits)

    link_cfg = LinkConfig.depolarise_config(
        prob_max_mixed, attempt_success_prob, attempt_duration, state_delay
    )
    link_between_cfg = LinkBetweenNodesConfig(
        node_id1=alice_id, node_id2=bob_id, link_config=link_cfg
    )
    network_cfg = ProcNodeNetworkConfig(
        nodes=[alice_node_cfg, bob_node_cfg], links=[link_between_cfg]
    )

    alice_program = load_program(alice_file)
    bob_program = load_program(bob_file)

    if num_pairs is not None:
        alice_input = ProgramInput({"bob_id": bob_id, "N": num_pairs})
        bob_input = ProgramInput({"alice_id": alice_id, "N": num_pairs})
    else:
        alice_input = ProgramInput({"bob_id": bob_id})
        bob_input = ProgramInput({"alice_id": alice_id})

    app_result = run_two_node_app(
        num_iterations=num_iterations,
        programs={"alice": alice_program, "bob": bob_program},
        program_inputs={"alice": alice_input, "bob": bob_input},
        network_cfg=network_cfg,
        linear=True,
    )

    alice_result = app_result.batch_results["alice"]
    bob_result = app_result.batch_results["bob"]

    return QkdResult(alice_result, bob_result)


def test_qkd_md_1pair():
    ns.sim_reset()

    # LogManager.set_log_level("INFO")

    num_iterations = 1000

    alice_file = "qkd_1pair_MD_alice.iqoala"
    bob_file = "qkd_1pair_MD_bob.iqoala"

    qkd_result = run_qkd(
        num_iterations,
        alice_file,
        bob_file,
        prob_max_mixed=0.4,
        attempt_success_prob=0.001,
        attempt_duration=1000,
        state_delay=0.0,
    )
    assert len(qkd_result.alice_result.results) == num_iterations
    assert len(qkd_result.bob_result.results) == num_iterations

    count_equal_outcomes = 0
    durations = []
    for i in range(num_iterations):
        alice_result = qkd_result.alice_result.results[i].values
        bob_result = qkd_result.bob_result.results[i].values
        if alice_result["m0"] == bob_result["m0"]:
            count_equal_outcomes += 1

        alice_start, alice_end = qkd_result.alice_result.timestamps[i]
        bob_start, bob_end = qkd_result.alice_result.timestamps[i]
        duration = max(alice_end, bob_end) - min(alice_start, bob_start)
        durations.append(duration)

    avg_duration = sum(durations) / len(durations)

    # We used a link that produced the state 0.4 * <maximally mixed> + 0.6 * Phi+.
    # Hence we expect the ratio of pairs with equal outcomes to be
    # 0.5 * 0.4                     +    1.0 * 0.6                   = 0.8
    # (mixed state -> 50% success)       (Phi+ -> 100% success)
    assert (count_equal_outcomes / num_iterations) <= 0.85
    assert (count_equal_outcomes / num_iterations) >= 0.75

    # On average, 1000 attempts were needed per EPR pair, which took 1000 ns each.
    # The number of attempts hence follows a geometric distribution.
    # The number of attempts, averaged over all 1000 iterations, should be within
    # the range (918, 1081) with probability 0.99.
    print(avg_duration)
    assert avg_duration > 918 * 1000
    assert avg_duration < 1081 * 1000


def test_qkd_md_npairs():
    ns.sim_reset()

    # LogManager.set_log_level("INFO")

    num_iterations = 1

    alice_file = "qkd_npairs_MD_alice.iqoala"
    bob_file = "qkd_npairs_MD_bob.iqoala"

    qkd_result = run_qkd(
        num_iterations,
        alice_file,
        bob_file,
        prob_max_mixed=0.4,
        attempt_success_prob=0.001,
        attempt_duration=1000,
        state_delay=0.0,
        num_pairs=1000,
    )
    assert len(qkd_result.alice_result.results) == num_iterations
    assert len(qkd_result.bob_result.results) == num_iterations

    count_equal_outcomes = 0
    durations = []
    for i in range(num_iterations):
        alice_outcomes = qkd_result.alice_result.results[i].values["outcomes"]
        bob_outcomes = qkd_result.bob_result.results[i].values["outcomes"]
        for alice, bob in zip(alice_outcomes, bob_outcomes):
            if alice == bob:
                count_equal_outcomes += 1

        alice_start, alice_end = qkd_result.alice_result.timestamps[i]
        bob_start, bob_end = qkd_result.alice_result.timestamps[i]
        duration = max(alice_end, bob_end) - min(alice_start, bob_start)
        durations.append(duration)

    avg_duration = sum(durations) / len(durations)

    # We used a link that produced the state 0.4 * <maximally mixed> + 0.6 * Phi+.
    # Hence we expect the ratio of pairs with equal outcomes to be
    # 0.5 * 0.4                     +    1.0 * 0.6                   = 0.8
    # (mixed state -> 50% success)       (Phi+ -> 100% success)
    assert (count_equal_outcomes / (num_iterations * 1000)) <= 0.85
    assert (count_equal_outcomes / (num_iterations * 1000)) >= 0.75

    # On average, 1000 attempts were needed per EPR pair, which took 1000 ns each.
    # The number of attempts hence follows a geometric distribution.
    # The number of attempts, averaged over all 1000 pairs per program,
    # should be within the range (918, 1081) with probability 0.99.
    print(avg_duration)
    assert avg_duration > 918 * 1000 * 1000
    assert avg_duration < 1081 * 1000 * 1000


if __name__ == "__main__":
    test_qkd_md_1pair()
    test_qkd_md_npairs()



from pydynaa import EventType

EVENT_WAIT = EventType("SCHEDULER_WAIT", "scheduler wait")
EPR_DELIVERY = EventType("EPR_DELIVERY", "EPR delivery")


# Signals inside a single node
SIGNAL_HOST_HOST_MSG = "EvHostHostMsg"
SIGNAL_HOST_QNOS_MSG = "EvHostQnosMsg"
SIGNAL_HOST_NSTK_MSG = "EvHostNstkMsg"
SIGNAL_QNOS_HOST_MSG = "EvQnosHostMsg"
SIGNAL_QNOS_NSTK_MSG = "EvQnosNstkMsg"
SIGNAL_NSTK_HOST_MSG = "EvNstkHostMsg"
SIGNAL_NSTK_QNOS_MSG = "EvNstkQnosMsg"
SIGNAL_NSTK_NSTK_MSG = "EvNstkNstkMsg"

# Signals between a node and the EntDist
SIGNAL_NSTK_ENTD_MSG = "EvNstkEntdMsg"
SIGNAL_ENTD_NSTK_MSG = "EvEntdNstkMsg"

# Global signals
SIGNAL_MEMORY_FREED = "EvMemoryFreed"
SIGNAL_TASK_COMPLETED = "TaskCompleted"
MSG_REQUEST_DELIVERED = "RequestDelivered"



from __future__ import annotations

import datetime
import json
import math
import os
import time
from argparse import ArgumentParser
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import List

import netsquid as ns

from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    ClassicalConnectionConfig,
    LatenciesConfig,
    NtfConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.program import BatchResult, ProgramInput
from qoala.util.runner import run_two_node_app_separate_inputs


def relative_to_cwd(file: str) -> str:
    return os.path.join(os.path.dirname(__file__), file)


def create_procnode_cfg(name: str, id: int, t1: int, t2: int) -> ProcNodeConfig:
    return ProcNodeConfig(
        node_name=name,
        node_id=id,
        topology=TopologyConfig.uniform_t1t2_qubits_perfect_gates_default_params(
            3, t1, t2
        ),
        latencies=LatenciesConfig(qnos_instr_time=1000),
        ntf=NtfConfig.from_cls_name("GenericNtf"),
        determ_sched=True,
    )


def load_program(path: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()
    return QoalaParser(text).parse()


@dataclass
class RemoteMbqcResult:
    client_results: BatchResult
    server_results: BatchResult


def run_remote_mbqc(
    num_iterations: int,
    theta0: float,
    theta1: float,
    theta2: float,
    naive: bool,
    t1: int,
    t2: int,
    cc: float,
) -> RemoteMbqcResult:
    ns.sim_reset()

    client_id = 1
    server_id = 0

    client_node_cfg = create_procnode_cfg("client", client_id, t1, t2)
    server_node_cfg = create_procnode_cfg("server", server_id, t1, t2)

    cconn = ClassicalConnectionConfig.from_nodes(client_id, server_id, cc)
    network_cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
        nodes=[client_node_cfg, server_node_cfg], link_duration=1000
    )
    network_cfg.cconns = [cconn]

    if naive:
        client_program = load_program("remote_mbqc_naive_client.iqoala")
        server_program = load_program("remote_mbqc_naive_server.iqoala")
    else:
        client_program = load_program("remote_mbqc_opt_client.iqoala")
        server_program = load_program("remote_mbqc_opt_server.iqoala")

    theta0_int = int(theta0 * 16 / math.pi)
    theta1_int = int(theta1 * 16 / math.pi)
    theta2_int = int(theta2 * 16 / math.pi)
    client_inputs = [
        ProgramInput(
            {
                "server_id": server_id,
                "theta0": theta0_int,
                "theta1": theta1_int,
                "theta2": theta2_int,
            }
        )
        for _ in range(num_iterations)
    ]
    server_inputs = [
        ProgramInput({"client_id": client_id}) for _ in range(num_iterations)
    ]

    app_result = run_two_node_app_separate_inputs(
        num_iterations=num_iterations,
        programs={"client": client_program, "server": server_program},
        program_inputs={"client": client_inputs, "server": server_inputs},
        network_cfg=network_cfg,
        linear=True,
    )

    client_result = app_result.batch_results["client"]
    server_result = app_result.batch_results["server"]

    return RemoteMbqcResult(client_result, server_result)


@dataclass
class DataPoint:
    naive: bool
    succ_prob: float


@dataclass
class DataMeta:
    timestamp: str
    num_iterations: int
    theta0: float
    theta1: float
    theta2: float
    t1: float
    t2: float
    cc: float
    sim_duration: float


@dataclass
class Data:
    meta: DataMeta
    data_points: List[DataPoint]


def remote_mbqc(
    naive: bool,
    num_iterations: int,
    t1: float,
    t2: float,
    cc: float,
    theta0: float,
    theta1: float,
    theta2: float,
) -> float:
    result = run_remote_mbqc(num_iterations, theta0, theta1, theta2, naive, t1, t2, cc)
    program_results = result.client_results.results
    # print(program_results)
    m0s = [result.values["m0"] for result in program_results]
    m1s = [result.values["m1"] for result in program_results]
    m2s = [result.values["m2"] for result in program_results]
    # print(m0s)
    # print(m1s)
    # print(m2s)

    successes = 0
    for m0, m1, m2 in zip(m0s, m1s, m2s):
        if m2 == (m0 == m1):
            successes += 1
    succ_prob = round(successes / num_iterations, 3)
    print(f"succ prob: {succ_prob}")
    return succ_prob


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--num_iterations", "-n", type=int, required=True)

    args = parser.parse_args()
    num_iterations = args.num_iterations

    t1 = 1e10
    t2 = 1e6
    cc = 1e5

    theta0 = math.pi / 2
    theta1 = 0
    theta2 = math.pi / 2

    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")

    start = time.time()
    data_points: List[DataPoint] = []

    succ_prob_naive = remote_mbqc(
        naive=True,
        num_iterations=num_iterations,
        t1=t1,
        t2=t2,
        cc=cc,
        theta0=theta0,
        theta1=theta1,
        theta2=theta2,
    )
    data_points.append(DataPoint(naive=True, succ_prob=succ_prob_naive))

    succ_prob_qoala = remote_mbqc(
        naive=False,
        num_iterations=num_iterations,
        t1=t1,
        t2=t2,
        cc=cc,
        theta0=theta0,
        theta1=theta1,
        theta2=theta2,
    )

    data_points.append(DataPoint(naive=False, succ_prob=succ_prob_qoala))

    end = time.time()
    duration = round(end - start, 2)

    abs_dir = relative_to_cwd(f"data")
    Path(abs_dir).mkdir(parents=True, exist_ok=True)
    last_path = os.path.join(abs_dir, "LAST.json")
    timestamp_path = os.path.join(abs_dir, f"{timestamp}.json")

    meta = DataMeta(
        timestamp=timestamp,
        num_iterations=num_iterations,
        theta0=theta0,
        theta1=theta1,
        theta2=theta2,
        t1=t1,
        t2=t2,
        cc=cc,
        sim_duration=duration,
    )
    data = Data(meta=meta, data_points=data_points)
    json_data = asdict(data)

    with open(last_path, "w") as datafile:
        json.dump(json_data, datafile)
    with open(timestamp_path, "w") as datafile:
        json.dump(json_data, datafile)



import json
import os
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import List

import dacite
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker


@dataclass
class DataPoint:
    sched_typ: str
    busy_factor: float
    busy_duration: float
    succ_prob: float
    succ_prob_lower: float
    succ_prob_upper: float
    makespan: float


@dataclass
class DataMeta:
    timestamp: str
    sim_duration: float
    num_iterations: int
    t1: float
    t2: float
    latency_factor: float
    num_const_tasks: int
    const_rate_factor: float
    cc_latency: float
    const_period: float
    const_start: float
    busy_factors: List[float]


@dataclass
class Data:
    meta: DataMeta
    data_points: List[DataPoint]


def relative_to_cwd(file: str) -> str:
    return os.path.join(os.path.dirname(__file__), file)


def create_png(filename: str):
    output_dir = relative_to_cwd("plots")
    Path(output_dir).mkdir(parents=True, exist_ok=True)
    output_path = os.path.join(output_dir, f"{filename}.png")
    plt.savefig(output_path)
    print(f"plot written to {output_path}")


def create_meta(filename: str, no_sched: Data, fcfs: Data, qoala: Data):
    output_dir = relative_to_cwd("plots")
    Path(output_dir).mkdir(parents=True, exist_ok=True)
    output_path = os.path.join(output_dir, f"{filename}.json")
    meta = {}
    meta["no_sched"] = no_sched.meta.timestamp
    meta["fcfs"] = fcfs.meta.timestamp
    meta["qoala"] = qoala.meta.timestamp
    with open(output_path, "w") as metafile:
        json.dump(meta, metafile)


def load_data(path: str) -> Data:
    with open(relative_to_cwd(path), "r") as f:
        all_data = json.load(f)

    return dacite.from_dict(Data, all_data)


def sched_type_to_label(sched_type: str) -> str:
    return {"NO_SCHED": "Baseline", "FCFS": "FCFS", "QOALA": "EDF"}[sched_type]


FORMATS = {
    "NO_SCHED": "^-",
    "FCFS": "s--",
    "QOALA": "D-.",
}

COLORS = {
    "NO_SCHED": "#1F77B4",
    "FCFS": "#2CA02C",
    "QOALA": "#FF7F0E",
}


def plot_sweep_busy_factor(
    timestamp: str, no_sched_data: Data, fcfs_data: Data, qoala_data: Data
) -> None:
    fig, ax = plt.subplots()

    ax.grid()
    ax.set_xlabel(
        "Classical task duration as fraction of classical communication latency",
        fontsize=12,
    )
    ax.set_ylabel("Success prob. of interactive quantum program", fontsize=12)
    ax.set_xscale("log")
    ax.tick_params(labelsize=11)

    ax2 = ax.twinx()
    ax2.set_ylabel("Makespan improvement factor (EDF)", fontsize=12)
    ax2.spines["right"].set_color("red")
    ax2.yaxis.label.set_color("red")
    ax2.tick_params(axis="y", colors="red", labelsize=11)

    lines = []

    for data in [no_sched_data, fcfs_data, qoala_data]:
        bf = [p.busy_factor for p in data.data_points]
        succ_probs = [p.succ_prob for p in data.data_points]
        error_plus = [p.succ_prob_upper - p.succ_prob for p in data.data_points]
        error_plus = [max(0, e) for e in error_plus]
        error_minus = [p.succ_prob - p.succ_prob_lower for p in data.data_points]
        error_minus = [max(0, e) for e in error_minus]
        errors = [error_minus, error_plus]
        sched_typ = data.data_points[0].sched_typ
        line = ax.errorbar(
            x=bf,
            y=succ_probs,
            yerr=errors,
            fmt=FORMATS[sched_typ],
            color=COLORS[sched_typ],
            label=sched_type_to_label(sched_typ),
        )

        lines.append(line)

    makespan_improvements = [
        p1.makespan / p2.makespan
        for p1, p2 in zip(no_sched_data.data_points, qoala_data.data_points)
    ]
    bf = [p.busy_factor for p in no_sched_data.data_points]
    print(bf)
    lines.append(
        ax2.errorbar(
            x=bf,
            y=makespan_improvements,
            # yerr=errors,
            fmt="o:r",
            label="Makespan improv.",
        )
    )

    def format_func(value, tick_number):
        # return f"{10.0**value:.1f}"
        return f"{value}"

    # Create a FuncFormatter object using the format function
    formatter = ticker.FuncFormatter(format_func)

    # Set the x-axis formatter
    ax.xaxis.set_major_formatter(formatter)
    ax.xaxis.set_major_locator(ticker.LogLocator(base=10.0))
    ax.xaxis.set_minor_locator(ticker.FixedLocator([0.1, 1, 10]))

    # ax2.set_yscale("log")
    labels = [line.get_label() for line in lines]

    ax.legend(lines, labels, loc="lower left", fontsize=11)

    # ax.set_title(
    #     "Success probability and makespan for quantum program in presence of busy classical program",
    #     wrap=True,
    # )

    # ax.set_ylim(0.5, 1.0)

    create_png("LAST")
    create_png(timestamp)


def sweep_busy_factor():
    no_sched_data = load_data("data/no_sched/LAST.json")
    fcfs_data = load_data("data/fcfs/LAST.json")
    qoala_data = load_data("data/qoala/LAST.json")

    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")

    create_meta("LAST_meta", no_sched_data, fcfs_data, qoala_data)
    create_meta(f"{timestamp}_meta", no_sched_data, fcfs_data, qoala_data)
    plot_sweep_busy_factor(timestamp, no_sched_data, fcfs_data, qoala_data)


if __name__ == "__main__":
    sweep_busy_factor()



from qoala.lang.ehi import EhiBuilder, UnitModule
from qoala.lang.parse import LocalRoutineParser, RequestRoutineParser
from qoala.lang.program import ProgramMeta, QoalaProgram
from qoala.lang.request import RequestRoutine
from qoala.lang.routine import LocalRoutine
from qoala.runtime.memory import ProgramMemory
from qoala.runtime.program import ProgramInput, ProgramInstance, ProgramResult
from qoala.sim.process import QoalaProcess


def create_local_routine() -> LocalRoutine:
    text = """
SUBROUTINE subrt1
    params: my_value
    returns: m
    uses: 
    keeps:
    request: 
  NETQASM_START
    set Q0 {my_value}
  NETQASM_END
    """

    parsed = LocalRoutineParser(text).parse()
    return parsed["subrt1"]


def create_request_routine() -> RequestRoutine:
    text = """
    REQUEST req1
      callback_type: wait_all
      callback: 
      return_vars: 
      remote_id: 1
      epr_socket_id: 0
      num_pairs: 3
      virt_ids: custom 1, 2, 3
      timeout: 1000
      fidelity: 0.65
      typ: measure_directly
      role: receive
        """
    return RequestRoutineParser(text).parse()["req1"]


def create_process(program: QoalaProgram) -> QoalaProcess:
    ehi = EhiBuilder.perfect_uniform(1, None, [], 0, [], 0)
    unit_module = UnitModule.from_full_ehi(ehi)

    instance = ProgramInstance(
        pid=0,
        program=program,
        inputs=ProgramInput({}),
        unit_module=unit_module,
    )
    mem = ProgramMemory(pid=0)

    process = QoalaProcess(
        prog_instance=instance,
        prog_memory=mem,
        csockets={},
        epr_sockets=program.meta.epr_sockets,
        result=ProgramResult(values={}),
    )
    return process


def test1():
    routine = create_local_routine()

    program = QoalaProgram(
        blocks=[], local_routines={"subrt1": routine}, meta=ProgramMeta.empty("")
    )
    process = create_process(program)

    assert len(process.get_all_local_routines()) == 1
    assert process.get_local_routine("subrt1") == routine
    assert len(process.qnos_mem.get_all_running_local_routines()) == 0


def test2():
    routine = create_request_routine()

    program = QoalaProgram(
        blocks=[], request_routines={"req1": routine}, meta=ProgramMeta.empty("")
    )

    process = create_process(program)
    assert len(process.get_all_request_routines()) == 1
    assert process.get_request_routine("req1") == routine
    assert len(process.qnos_mem.get_all_running_request_routines()) == 0


if __name__ == "__main__":
    test1()
    test2()



from __future__ import annotations

import logging
from typing import Dict, Generator, List, Optional, Union

from netqasm.lang.operand import Template

from pydynaa import EventExpression
from qoala.lang import hostlang
from qoala.lang.hostlang import ClassicalIqoalaOp, IqoalaSingleton, IqoalaVectorElement
from qoala.lang.request import CallbackType
from qoala.runtime.memory import HostMemory
from qoala.runtime.message import LrCallTuple, RrCallTuple
from qoala.runtime.sharedmem import MemAddr
from qoala.sim.host.hostinterface import HostInterface, HostLatencies
from qoala.sim.process import QoalaProcess
from qoala.util.logging import LogManager


class HostProcessor:
    """Does not have state itself. Acts on and changes process objects."""

    def __init__(
        self,
        interface: HostInterface,
        latencies: HostLatencies,
        asynchronous: bool = False,
    ) -> None:
        self._interface = interface
        self._latencies = latencies
        self._asynchronous = asynchronous

        # TODO: name
        self._name = f"{interface.name}_HostProcessor"
        self._logger: logging.Logger = LogManager.get_stack_logger(  # type: ignore
            f"{self.__class__.__name__}({self._name})"
        )

    def initialize(self, process: QoalaProcess) -> None:
        host_mem = process.prog_memory.host_mem
        inputs = process.prog_instance.inputs
        for name, value in inputs.values.items():
            host_mem.write(name, value)

    @staticmethod
    def _read_value_from_host_mem(
        arg: Union[IqoalaSingleton, IqoalaVectorElement], host_mem: HostMemory
    ) -> int:
        if isinstance(arg, hostlang.IqoalaSingleton):
            return host_mem.read(arg.name)
        else:
            loc = arg.name
            index = arg.index
            return host_mem.read_vec(loc)[index]

    def assign_instr_index(
        self, process: QoalaProcess, instr_idx: int
    ) -> Generator[EventExpression, None, None]:
        program = process.prog_instance.program
        instr = program.instructions[instr_idx]
        yield from self.assign_instr(process, instr)

    def assign_block(
        self, process: QoalaProcess, block_name: str
    ) -> Generator[EventExpression, None, None]:
        block = process.program.get_block(block_name)

        for instr in block.instructions:
            yield from self.assign_instr(process, instr)
            if self._interface.program_instance_jumps[process.pid] != -1:
                break

    def assign_instr(
        self, process: QoalaProcess, instr: hostlang.ClassicalIqoalaOp
    ) -> Generator[EventExpression, None, None]:
        csockets = process.csockets
        host_mem = process.prog_memory.host_mem
        pid = process.pid

        # Instruction duration is simulated for each instruction by adding a "wait".
        # Duration of wait is "host_instr_time".
        # Half of it is applied *before* any operations.
        # The other half is applied *after* all 'reads' from shared memory,
        # and *before* any 'writes' to shared memory.
        # See #29 for rationale.

        # Apply half of the instruction duration.
        instr_time = self._latencies.host_instr_time
        first_half = instr_time / 2
        second_half = instr_time - first_half  # just to make it adds up
        self._interface.program_instance_jumps[pid] = -1
        self._logger.debug(f"Interpreting LHR instruction {instr}")
        if isinstance(instr, hostlang.AssignCValueOp):
            yield from self._interface.wait(first_half)
            value = instr.attributes[0]
            assert isinstance(value, int)
            assert isinstance(instr.results, hostlang.IqoalaSingleton)
            loc = instr.results.name  # type: ignore
            self._logger.debug(f"writing {value} to {loc}")
            yield from self._interface.wait(second_half)
            host_mem.write(loc, value)
        elif isinstance(instr, hostlang.BusyOp):
            value = instr.attributes[0]
            if not isinstance(value, int):
                value = host_mem.read(value)
            self._logger.debug(f"busy for {value} ns")
            yield from self._interface.wait(value)
        elif isinstance(instr, hostlang.SendCMsgOp):
            assert isinstance(
                instr.arguments[0], hostlang.IqoalaSingleton
            ) or isinstance(instr.arguments[0], hostlang.IqoalaVectorElement)
            assert isinstance(
                instr.arguments[1], hostlang.IqoalaSingleton
            ) or isinstance(instr.arguments[1], hostlang.IqoalaVectorElement)

            csck_id = self._read_value_from_host_mem(instr.arguments[0], host_mem)

            csck = csockets[csck_id]
            if isinstance(instr.arguments[1], hostlang.IqoalaSingleton):
                value = host_mem.read(instr.arguments[1].name)
            else:
                loc = instr.arguments[1].name
                index = instr.arguments[1].index
                value = host_mem.read_vec(loc)[index]
            self._logger.info(f"sending msg {value}")
            csck.send_int(value)
            # Simulate instruction duration.
            yield from self._interface.wait(self._latencies.host_instr_time)
        elif isinstance(instr, hostlang.ReceiveCMsgOp):
            assert isinstance(
                instr.arguments[0], hostlang.IqoalaSingleton
            ) or isinstance(instr.arguments[0], hostlang.IqoalaVectorElement)
            assert isinstance(instr.results, hostlang.IqoalaSingleton)

            csck_id = self._read_value_from_host_mem(instr.arguments[0], host_mem)

            csck = csockets[csck_id]

            # TODO: refactor
            tup = (csck.remote_pid, pid)
            if tup not in self._interface.get_available_messages(csck.remote_name):
                msg = yield from csck.recv_int()
            else:
                msg = csck.read_int()

            yield from self._interface.wait(self._latencies.host_peer_latency)
            host_mem.write(instr.results.name, msg)
            self._logger.info(f"received msg {msg}")
        elif isinstance(instr, hostlang.AddCValueOp):
            yield from self._interface.wait(first_half)
            assert isinstance(
                instr.arguments[0], hostlang.IqoalaSingleton
            ) or isinstance(instr.arguments[0], hostlang.IqoalaVectorElement)
            assert isinstance(
                instr.arguments[1], hostlang.IqoalaSingleton
            ) or isinstance(instr.arguments[1], hostlang.IqoalaVectorElement)

            arg0 = self._read_value_from_host_mem(instr.arguments[0], host_mem)
            arg1 = self._read_value_from_host_mem(instr.arguments[1], host_mem)

            assert isinstance(instr.results, hostlang.IqoalaSingleton)
            loc = instr.results.name  # type: ignore
            result = arg0 + arg1
            self._logger.debug(f"computing {loc} = {arg0} + {arg1} = {result}")
            # Simulate instruction duration.
            yield from self._interface.wait(second_half)
            host_mem.write(loc, result)
        elif isinstance(instr, hostlang.MultiplyConstantCValueOp):
            yield from self._interface.wait(first_half)
            assert isinstance(
                instr.arguments[0], hostlang.IqoalaSingleton
            ) or isinstance(instr.arguments[0], hostlang.IqoalaVectorElement)

            arg0 = self._read_value_from_host_mem(instr.arguments[0], host_mem)

            const = instr.attributes[0]
            assert isinstance(const, int)
            assert isinstance(instr.results, hostlang.IqoalaSingleton)
            loc = instr.results.name  # type: ignore
            result = arg0 * const
            self._logger.debug(f"computing {loc} = {arg0} * {const} = {result}")
            # Simulate instruction duration.
            yield from self._interface.wait(second_half)
            host_mem.write(loc, result)
        elif isinstance(instr, hostlang.BitConditionalMultiplyConstantCValueOp):
            yield from self._interface.wait(first_half)

            assert isinstance(
                instr.arguments[0], hostlang.IqoalaSingleton
            ) or isinstance(instr.arguments[0], hostlang.IqoalaVectorElement)
            assert isinstance(
                instr.arguments[1], hostlang.IqoalaSingleton
            ) or isinstance(instr.arguments[1], hostlang.IqoalaVectorElement)

            arg0 = self._read_value_from_host_mem(instr.arguments[0], host_mem)
            cond = self._read_value_from_host_mem(instr.arguments[1], host_mem)

            const = instr.attributes[0]
            assert isinstance(const, int)
            assert isinstance(instr.results, hostlang.IqoalaSingleton)
            loc = instr.results.name  # type: ignore
            if cond == 1:
                result = arg0 * const
            else:
                result = arg0
            self._logger.debug(f"computing {loc} = {arg0} * {const}^{cond} = {result}")
            # Simulate instruction duration.
            yield from self._interface.wait(second_half)
            host_mem.write(loc, result)
        elif isinstance(instr, hostlang.ReturnResultOp):
            yield from self._interface.wait(first_half)
            assert isinstance(
                instr.arguments[0], hostlang.IqoalaSingleton
            ) or isinstance(instr.arguments[0], hostlang.IqoalaVector)
            loc = instr.arguments[0].name
            if isinstance(instr.arguments[0], hostlang.IqoalaSingleton):
                value = host_mem.read(loc)
            else:
                value = host_mem.read_vec(loc)
            self._logger.debug(f"returning {loc} = {value}")
            # Simulate instruction duration.
            yield from self._interface.wait(second_half)
            process.result.values[loc] = value
        elif isinstance(instr, hostlang.JumpOp):
            yield from self._interface.wait(first_half)
            assert isinstance(instr.attributes[0], str)
            block_name = instr.attributes[0]
            self._logger.debug(f"jumping to block {block_name}")
            block_id = process.prog_instance.program.get_block_id(block_name)
            self._interface.program_instance_jumps[pid] = block_id
            yield from self._interface.wait(second_half)
        elif isinstance(instr, hostlang.BranchIfEqualOp):
            yield from self._branch(process, instr, "__eq__")
        elif isinstance(instr, hostlang.BranchIfNotEqualOp):
            yield from self._branch(process, instr, "__ne__")
        elif isinstance(instr, hostlang.BranchIfLessThanOp):
            yield from self._branch(process, instr, "__lt__")
        elif isinstance(instr, hostlang.BranchIfGreaterThanOp):
            yield from self._branch(process, instr, "__gt__")

    def _branch(
        self, process: QoalaProcess, instr: ClassicalIqoalaOp, comparison_op: str
    ) -> Generator[EventExpression, None, None]:
        instr_time = self._latencies.host_instr_time
        first_half = instr_time / 2
        second_half = instr_time - first_half  # just to make it adds up
        yield from self._interface.wait(first_half)

        assert isinstance(instr.arguments[0], hostlang.IqoalaSingleton) or isinstance(
            instr.arguments[0], hostlang.IqoalaVectorElement
        )
        assert isinstance(instr.arguments[1], hostlang.IqoalaSingleton) or isinstance(
            instr.arguments[1], hostlang.IqoalaVectorElement
        )

        host_mem = process.prog_memory.host_mem
        pid = process.pid

        value0 = self._read_value_from_host_mem(instr.arguments[0], host_mem)
        value1 = self._read_value_from_host_mem(instr.arguments[1], host_mem)

        assert isinstance(instr.attributes[0], str)
        block_name = instr.attributes[0]
        if getattr(value0, comparison_op)(value1):
            block_id = process.prog_instance.program.get_block_id(block_name)
            self._interface.program_instance_jumps[pid] = block_id
        yield from self._interface.wait(second_half)

    def prepare_lr_call(
        self, process: QoalaProcess, instr: hostlang.RunSubroutineOp
    ) -> LrCallTuple:
        host_mem = process.prog_memory.host_mem

        assert isinstance(instr.arguments[0], hostlang.IqoalaTuple)
        arg_vec: hostlang.IqoalaTuple = instr.arguments[0]
        args = arg_vec.values
        subrt_name = instr.attributes[0]
        assert isinstance(subrt_name, str)

        routine = process.get_local_routine(subrt_name)
        self._logger.info(f"executing subroutine {routine.name}")
        self._logger.debug(f"subroutine contents of {routine.name}: \n{routine}")

        arg_values = {arg: host_mem.read(arg) for arg in args}

        # self._logger.info(f"instantiating subroutine with values {arg_values}")
        # process.instantiate_routine(subrt_name, arg_values)

        shared_mem = process.prog_memory.shared_mem

        # Allocate input memory and write args to it.
        input_addr = shared_mem.allocate_lr_in(len(arg_values))
        shared_mem.write_lr_in(input_addr, list(arg_values.values()))

        # Check if the expected number of results match the number of result variables.
        result_vars_len = self._calculate_result_size(instr.results)
        assert result_vars_len == routine.get_return_size()

        # Allocate result memory.
        result_addr = shared_mem.allocate_lr_out(result_vars_len)

        return LrCallTuple(subrt_name, input_addr, result_addr)

    def post_lr_call(
        self,
        process: QoalaProcess,
        instr: hostlang.RunSubroutineOp,
        lrcall: LrCallTuple,
    ) -> None:
        shared_mem = process.prog_memory.shared_mem

        # Read the results from shared memory.
        result_vars_len = self._calculate_result_size(instr.results)
        result = shared_mem.read_lr_out(lrcall.result_addr, result_vars_len)
        assert len(result) == result_vars_len

        # Copy results to local host variables.
        if isinstance(instr.results, hostlang.IqoalaSingleton):
            for value, var in zip(result, instr.results.name):
                process.host_mem.write(var, value)
        elif isinstance(instr.results, hostlang.IqoalaTuple):
            result_tup: hostlang.IqoalaTuple = instr.results
            for value, var in zip(result, result_tup.values):
                process.host_mem.write(var, value)
        elif isinstance(instr.results, hostlang.IqoalaVector):
            result_vec: hostlang.IqoalaVector = instr.results
            process.host_mem.write_vec(result_vec.name, result)
        elif instr.results is None:
            pass
        else:
            raise RuntimeError

    def prepare_rr_call(
        self, process: QoalaProcess, instr: hostlang.RunRequestOp
    ) -> RrCallTuple:
        host_mem = process.prog_memory.host_mem

        assert isinstance(instr.arguments[0], hostlang.IqoalaTuple)
        arg_vec: hostlang.IqoalaTuple = instr.arguments[0]
        args = arg_vec.values
        routine_name = instr.attributes[0]
        assert isinstance(routine_name, str)

        routine = process.get_request_routine(routine_name)
        self._logger.info(f"executing request routine {routine}")

        arg_values = {arg: host_mem.read(arg) for arg in args}

        shared_mem = process.prog_memory.shared_mem

        # Allocate input memory for RR itself and write args to it.
        input_addr = shared_mem.allocate_rr_in(len(arg_values))
        shared_mem.write_rr_in(input_addr, list(arg_values.values()))

        cb_input_addrs: List[MemAddr] = []
        cb_output_addrs: List[MemAddr] = []

        # TODO: refactor!!
        # the `num_pairs` entry of an RR may be a template.
        # Its value should be provided in the ProgramInput of this ProgamInstance.
        # This value is filled in when `instantiating` the RR, but currently this only
        # happens by the NetstackProcessor when it's assigned to execute the RR.
        # The filled-in value is then part of a `RunningRequestRoutine`. However, it is
        # not accessible by this code here.
        # For now we use the following 'hack' where we peek in the ProgramInputs:
        prog_input = process.prog_instance.inputs.values
        if isinstance(routine.request.num_pairs, Template):
            template_name = routine.request.num_pairs.name
            num_pairs = prog_input[template_name]
        else:
            num_pairs = routine.request.num_pairs

        # Calculate total return size (RR itself + callbacks).
        total_return_size = 0

        # Allocate memory for callbacks.
        if routine.callback is not None:
            if routine.callback_type == CallbackType.SEQUENTIAL:
                cb_routine = process.get_local_routine(routine.callback)
                for _ in range(num_pairs):
                    # Allocate input memory.
                    cb_args = cb_routine.subroutine.arguments
                    # TODO: can it just be LR_in instead of CR_in?
                    cb_input_addrs.append(shared_mem.allocate_cr_in(len(cb_args)))

                    # Allocate result memory.
                    cb_results_len = cb_routine.get_return_size()
                    cb_output_addrs.append(shared_mem.allocate_lr_out(cb_results_len))
                    total_return_size += cb_results_len
            elif routine.callback_type == CallbackType.WAIT_ALL:
                cb_routine = process.get_local_routine(routine.callback)
                # Allocate input memory.
                cb_args = cb_routine.subroutine.arguments
                # TODO: can it just be LR_in instead of CR_in?
                cb_input_addrs.append(shared_mem.allocate_cr_in(len(cb_args)))

                # Allocate result memory.
                cb_results_len = cb_routine.get_return_size()
                cb_output_addrs.append(shared_mem.allocate_lr_out(cb_results_len))
                total_return_size += cb_results_len

        # Allocate result memory for RR itself.
        routine_return_size = routine.get_return_size(prog_input)
        result_addr = shared_mem.allocate_rr_out(routine_return_size)
        total_return_size += routine_return_size

        # Check that host variables match the RR total result in length
        result_vars_len = self._calculate_result_size(instr.results, prog_input)
        assert result_vars_len == total_return_size

        return RrCallTuple(
            routine_name, input_addr, result_addr, cb_input_addrs, cb_output_addrs
        )

    def post_rr_call(
        self, process: QoalaProcess, instr: hostlang.RunRequestOp, rrcall: RrCallTuple
    ) -> None:
        shared_mem = process.prog_memory.shared_mem
        routine = process.get_request_routine(rrcall.routine_name)

        # Bit of a hack; see prepare_rr_call comments.
        prog_input = process.prog_instance.inputs.values
        if isinstance(routine.request.num_pairs, Template):
            template_name = routine.request.num_pairs.name
            num_pairs = prog_input[template_name]
        else:
            num_pairs = routine.request.num_pairs

        # Read the RR results from shared memory.
        rr_result = shared_mem.read_rr_out(
            rrcall.result_addr, routine.get_return_size(prog_input)
        )

        # Read the callback results from shared memory.
        cb_results: List[int] = []
        if routine.callback is not None:
            if routine.callback_type == CallbackType.SEQUENTIAL:
                cb_routine = process.get_local_routine(routine.callback)
                for i in range(num_pairs):
                    # Read result memory.
                    cb_results_len = cb_routine.get_return_size()
                    cb_output_addr = rrcall.cb_output_addrs[i]
                    result = shared_mem.read_lr_out(cb_output_addr, cb_results_len)
                    cb_results.extend(result)
            elif routine.callback_type == CallbackType.WAIT_ALL:
                cb_routine = process.get_local_routine(routine.callback)
                # Read result memory.
                cb_results_len = cb_routine.get_return_size()
                cb_output_addr = rrcall.cb_output_addrs[0]
                result = shared_mem.read_lr_out(cb_output_addr, cb_results_len)
                cb_results.extend(result)

        all_results = rr_result + cb_results
        # At this point, `all_results` contains all the results of both the RR itself
        # as well as from all the callbacks, in order.
        result_vars_len = self._calculate_result_size(instr.results, prog_input)
        assert len(all_results) == result_vars_len

        # Collect the host variables to which to copy these results.
        if isinstance(instr.results, hostlang.IqoalaSingleton):
            loc = instr.results.name
            value = all_results[0]
            process.host_mem.write(loc, value)
        elif isinstance(instr.results, hostlang.IqoalaTuple):
            result_tup: hostlang.IqoalaTuple = instr.results
            for value, var in zip(all_results, result_tup.values):
                process.host_mem.write(var, value)
        elif isinstance(instr.results, hostlang.IqoalaVector):
            result_vec: hostlang.IqoalaVector = instr.results
            process.host_mem.write_vec(result_vec.name, all_results)
        elif instr.results is None:
            pass
        else:
            raise RuntimeError

    def _calculate_result_size(
        self,
        results: hostlang.IqoalaVar,
        prog_input: Optional[Dict[str, int]] = None,
    ) -> int:
        if results is None:
            return 0
        elif isinstance(results, hostlang.IqoalaTuple):
            return len(results.values)
        elif isinstance(results, hostlang.IqoalaVector):
            if isinstance(results.size, int):
                return results.size
            else:  # Size is a variable. Get its value from the Program Inputs.
                assert prog_input is not None
                return prog_input[results.size]
        elif isinstance(results, hostlang.IqoalaSingleton):
            return 1
        else:
            raise RuntimeError



import os

from qoala.lang.hostlang import BasicBlockType
from qoala.lang.parse import QoalaParser
from qoala.runtime.lhi import (
    LhiLatencies,
    LhiLinkInfo,
    LhiNetworkInfo,
    LhiProcNodeInfo,
    LhiTopologyBuilder,
)
from qoala.runtime.ntf import GenericNtf
from qoala.runtime.task import (
    HostLocalTask,
    MultiPairCallbackTask,
    MultiPairTask,
    PostCallTask,
    PreCallTask,
    SinglePairCallbackTask,
    SinglePairTask,
    TaskDurationEstimator,
    TaskGraph,
    TaskGraphBuilder,
)
from qoala.sim.build import build_network_from_lhi
from qoala.sim.network import ProcNodeNetwork


def relative_path(path: str) -> str:
    return os.path.join(os.getcwd(), os.path.dirname(__file__), path)


CL = BasicBlockType.CL
CC = BasicBlockType.CC
QL = BasicBlockType.QL
QC = BasicBlockType.QC


def setup_network() -> ProcNodeNetwork:
    topology = LhiTopologyBuilder.perfect_uniform_default_gates(num_qubits=3)
    latencies = LhiLatencies(
        host_instr_time=1000, qnos_instr_time=2000, host_peer_latency=3000
    )
    link_info = LhiLinkInfo.perfect(duration=20_000)

    alice_lhi = LhiProcNodeInfo(
        name="alice", id=0, topology=topology, latencies=latencies
    )
    nodes = {0: "alice", 1: "bob"}
    network_lhi = LhiNetworkInfo.fully_connected(nodes, link_info)
    bob_lhi = LhiProcNodeInfo(name="bob", id=1, topology=topology, latencies=latencies)
    ntfs = [GenericNtf(), GenericNtf()]
    return build_network_from_lhi([alice_lhi, bob_lhi], ntfs, network_lhi)


def test_qoala_tasks_1_pair_callback():
    network = setup_network()
    alice = network.nodes["alice"]

    path = relative_path("test_callbacks_1_pair.iqoala")
    with open(path) as file:
        text = file.read()
    program = QoalaParser(text).parse()

    cb = program.local_routines["meas_1_pair"]

    cpu_time = alice.local_ehi.latencies.host_instr_time
    cb_time = TaskDurationEstimator.lr_duration(alice.local_ehi, cb)
    pair_time = alice.network_ehi.get_link(0, 1).duration

    pid = 3
    task_graph = TaskGraphBuilder.from_program(
        program, pid, alice.local_ehi, alice.network_ehi
    )

    expected_tasks = [
        # blk_1_pair_wait_all
        PreCallTask(0, pid, "blk_1_pair_wait_all", 0, cpu_time),
        PostCallTask(1, pid, "blk_1_pair_wait_all", 0, cpu_time),
        MultiPairTask(2, pid, 0, pair_time),
        MultiPairCallbackTask(3, pid, "meas_1_pair", 0, cb_time),
        # blk_1_pair_sequential
        PreCallTask(4, pid, "blk_1_pair_sequential", 4, cpu_time),
        PostCallTask(5, pid, "blk_1_pair_sequential", 4, cpu_time),
        SinglePairTask(6, pid, 0, 4, pair_time),
        SinglePairCallbackTask(7, pid, "meas_1_pair", 0, 4, cb_time),
    ]

    expected_precedences = [
        (0, 2),  # rr after precall
        (2, 3),  # callback after rr
        (3, 1),  # postcall after callback
        (1, 4),  # second block after first block
        (4, 6),  # rr after precall
        (6, 7),  # callback after rr
        (7, 5),  # postcall after callback
    ]

    expected_graph = TaskGraph()
    expected_graph.add_tasks(expected_tasks)
    expected_graph.add_precedences(expected_precedences)

    assert task_graph == expected_graph


def test_qoala_tasks_2_pairs_callback():
    network = setup_network()
    alice = network.nodes["alice"]

    path = relative_path("test_callbacks_2_pairs.iqoala")
    with open(path) as file:
        text = file.read()
    program = QoalaParser(text).parse()

    cb1 = program.local_routines["meas_1_pair"]
    cb2 = program.local_routines["meas_2_pairs"]
    cpu_time = alice.local_ehi.latencies.host_instr_time

    cb1_time = TaskDurationEstimator.lr_duration(alice.local_ehi, cb1)
    cb2_time = TaskDurationEstimator.lr_duration(alice.local_ehi, cb2)
    pair_time = alice.network_ehi.get_link(0, 1).duration

    pid = 3
    task_graph = TaskGraphBuilder.from_program(
        program, pid, alice.local_ehi, alice.network_ehi
    )

    expected_tasks = [
        # blk_2_pairs_wait_all
        PreCallTask(0, pid, "blk_2_pairs_wait_all", 0, cpu_time),
        PostCallTask(1, pid, "blk_2_pairs_wait_all", 0, cpu_time),
        MultiPairTask(2, pid, 0, 2 * pair_time),
        MultiPairCallbackTask(3, pid, "meas_2_pairs", 0, cb2_time),
        # blk_2_pairs_sequential
        PreCallTask(4, pid, "blk_2_pairs_sequential", 4, cpu_time),
        PostCallTask(5, pid, "blk_2_pairs_sequential", 4, cpu_time),
        SinglePairTask(6, pid, 0, 4, pair_time),
        SinglePairCallbackTask(7, pid, "meas_1_pair", 0, 4, cb1_time),
        SinglePairTask(8, pid, 1, 4, pair_time),
        SinglePairCallbackTask(9, pid, "meas_1_pair", 1, 4, cb1_time),
    ]

    expected_precedences = [
        (0, 2),  # rr after precall
        (2, 3),  # callback after rr
        (3, 1),  # postcall after callback
        (1, 4),  # second block after first block
        (4, 6),  # 1st pair after precall
        (6, 7),  # 1st pair callback after 1st pair rr
        (4, 8),  # 2nd pair after precall
        (8, 9),  # 2nd pair callback after 2nd pair rr
        (7, 5),  # postcall after 1st pair callback
        (9, 5),  # postcall after 2nd pair callback
    ]

    expected_graph = TaskGraph()
    expected_graph.add_tasks(expected_tasks)
    expected_graph.add_precedences(expected_precedences)

    assert task_graph == expected_graph


def test_deadlines():
    path = relative_path("test_deadlines.iqoala")
    with open(path) as file:
        text = file.read()
    program = QoalaParser(text).parse()

    pid = 3
    task_graph = TaskGraphBuilder.from_program(program, pid)

    expected_tasks = [
        HostLocalTask(0, pid, "b0"),
        HostLocalTask(1, pid, "b1"),
    ]
    expected_precedences = [(0, 1)]
    expected_deadlines = [((0, 1), 100)]

    expected_graph = TaskGraph()
    expected_graph.add_tasks(expected_tasks)
    expected_graph.add_precedences(expected_precedences)
    expected_graph.add_rel_deadlines(expected_deadlines)

    assert task_graph == expected_graph


if __name__ == "__main__":
    test_qoala_tasks_1_pair_callback()
    test_qoala_tasks_2_pairs_callback()
    test_deadlines()



import pytest
from netqasm.lang.instr import core, nv, trapped_ion
from netqasm.lang.instr.flavour import NVFlavour, TrappedIonFlavour
from netsquid.components.instructions import (
    INSTR_CXDIR,
    INSTR_INIT,
    INSTR_MEASURE,
    INSTR_ROT_X,
    INSTR_ROT_Y,
    INSTR_ROT_Z,
)

from qoala.lang.common import MultiQubit
from qoala.lang.ehi import EhiGateInfo, EhiQubitInfo
from qoala.runtime.config import DepolariseSamplerConfig, LinkConfig
from qoala.runtime.instructions import INSTR_ROT_X_ALL
from qoala.runtime.lhi import (
    LhiLatencies,
    LhiLinkInfo,
    LhiNetworkInfo,
    LhiTopologyBuilder,
)
from qoala.runtime.lhi_to_ehi import LhiConverter
from qoala.runtime.ntf import NvNtf, TrappedIonNtf
from qoala.util.math import prob_max_mixed_to_fidelity


def test_topology_to_ehi():
    topology = LhiTopologyBuilder.perfect_uniform(
        num_qubits=2,
        single_instructions=[
            INSTR_INIT,
            INSTR_ROT_X,
            INSTR_ROT_Y,
            INSTR_ROT_Z,
            INSTR_MEASURE,
        ],
        single_duration=5e3,
        two_instructions=[INSTR_CXDIR],
        two_duration=100e3,
    )

    latencies = LhiLatencies(
        host_instr_time=2,
        qnos_instr_time=3,
        host_peer_latency=4,
    )

    interface = NvNtf()
    ehi = LhiConverter.to_ehi(topology, interface, latencies)

    assert ehi.qubit_infos == {
        0: EhiQubitInfo(is_communication=True, decoherence_rate=0),
        1: EhiQubitInfo(is_communication=True, decoherence_rate=0),
    }

    assert ehi.flavour == NVFlavour

    single_gates = [
        EhiGateInfo(instr, 5e3, 0)
        for instr in [
            core.InitInstruction,
            nv.RotXInstruction,
            nv.RotYInstruction,
            nv.RotZInstruction,
            core.MeasInstruction,
        ]
    ]
    assert ehi.single_gate_infos == {0: single_gates, 1: single_gates}

    multi_gates = [EhiGateInfo(nv.ControlledRotXInstruction, 100e3, 0)]

    assert ehi.multi_gate_infos == {
        MultiQubit([0, 1]): multi_gates,
        MultiQubit([1, 0]): multi_gates,
    }

    assert ehi.latencies.host_instr_time == 2
    assert ehi.latencies.qnos_instr_time == 3
    assert ehi.latencies.host_peer_latency == 4


def test_topology_to_ehi_trapped_ion():
    topology = LhiTopologyBuilder.perfect_uniform(
        num_qubits=2,
        single_instructions=[
            INSTR_INIT,
            INSTR_ROT_Z,
            INSTR_MEASURE,
        ],
        single_duration=5e3,
        two_instructions=[],
        two_duration=0,
        all_qubit_instructions=[INSTR_ROT_X_ALL],
        all_qubit_duration=50e3,
    )

    latencies = LhiLatencies(
        host_instr_time=2,
        qnos_instr_time=3,
        host_peer_latency=4,
    )

    interface = TrappedIonNtf()
    ehi = LhiConverter.to_ehi(topology, interface, latencies)

    assert ehi.qubit_infos == {
        0: EhiQubitInfo(is_communication=True, decoherence_rate=0),
        1: EhiQubitInfo(is_communication=True, decoherence_rate=0),
    }

    assert ehi.flavour == TrappedIonFlavour

    single_gates = [
        EhiGateInfo(instr, 5e3, 0)
        for instr in [
            core.InitInstruction,
            trapped_ion.RotZInstruction,
            core.MeasInstruction,
        ]
    ]
    assert ehi.single_gate_infos == {0: single_gates, 1: single_gates}

    assert ehi.multi_gate_infos == {
        MultiQubit([0, 1]): [],
        MultiQubit([1, 0]): [],
    }

    all_qubit_gates = [EhiGateInfo(trapped_ion.AllQubitsRotXInstruction, 50e3, 0)]
    assert all_qubit_gates == ehi.all_qubit_gate_infos

    assert ehi.latencies.host_instr_time == 2
    assert ehi.latencies.qnos_instr_time == 3
    assert ehi.latencies.host_peer_latency == 4


def test_qubit_to_ehi():
    lhi_qubit = LhiTopologyBuilder.t1t2_qubit(is_communication=False, t1=3, t2=4)
    ehi_qubit = LhiConverter.qubit_info_to_ehi(lhi_qubit)

    assert not ehi_qubit.is_communication
    # currently it converts T1 into decoherence rate
    assert ehi_qubit.decoherence_rate == 3

    lhi_qubit2 = LhiTopologyBuilder.perfect_qubit(is_communication=True)
    ehi_qubit2 = LhiConverter.qubit_info_to_ehi(lhi_qubit2)

    assert ehi_qubit2.is_communication
    # currently it converts T1 into decoherence rate
    assert ehi_qubit2.decoherence_rate == 0


def test_gate_to_ehi():
    instructions = [INSTR_INIT, INSTR_ROT_X, INSTR_ROT_Y, INSTR_ROT_Z, INSTR_MEASURE]
    lhi_gates = LhiTopologyBuilder.perfect_gates(
        duration=5,
        instructions=instructions,
    )

    nvNtf = NvNtf()
    ehi_gates = [
        LhiConverter.gate_info_to_ehi(lhi_gate, nvNtf) for lhi_gate in lhi_gates
    ]

    assert ehi_gates == [
        EhiGateInfo(
            nvNtf.native_to_netqasm(lhi_gate.instruction)[0],
            lhi_gate.duration,
            0,
        )
        for lhi_gate in lhi_gates
    ]

    lhi_gates2 = LhiTopologyBuilder.depolar_gates(
        duration=5, instructions=instructions, depolar_rate=0.5
    )
    ehi_gates2 = [
        LhiConverter.gate_info_to_ehi(lhi_gate, nvNtf) for lhi_gate in lhi_gates2
    ]

    assert ehi_gates2 == [
        EhiGateInfo(
            nvNtf.native_to_netqasm(lhi_gate.instruction)[0],
            lhi_gate.duration,
            0.5,
        )
        for lhi_gate in lhi_gates2
    ]


def test_link_info_to_ehi_perfect():
    cfg = LinkConfig.perfect_config(state_delay=1200)
    lhi_info = LhiLinkInfo.from_config(cfg)
    ehi_info = LhiConverter.link_info_to_ehi(lhi_info)

    assert ehi_info.duration == 1200
    assert ehi_info.fidelity == 1.0


def test_link_info_to_ehi_depolarise():
    state_delay = 500
    cycle_time = 10
    prob_max_mixed = 0.3
    prob_success = 0.1

    cfg = LinkConfig(
        state_delay=state_delay,
        sampler_config_cls="DepolariseSamplerConfig",
        sampler_config=DepolariseSamplerConfig(
            cycle_time=cycle_time,
            prob_max_mixed=prob_max_mixed,
            prob_success=prob_success,
        ),
    )
    lhi_info = LhiLinkInfo.from_config(cfg)

    ehi_info = LhiConverter.link_info_to_ehi(lhi_info)

    expected_duration = (cycle_time / prob_success) + state_delay
    expected_fidelity = prob_max_mixed_to_fidelity(2, prob_max_mixed)
    assert ehi_info.duration == pytest.approx(expected_duration)
    assert ehi_info.fidelity == pytest.approx(expected_fidelity)


def test_network_to_ehi():
    depolar_link = LhiLinkInfo.depolarise(
        cycle_time=10, prob_max_mixed=0.2, prob_success=0.5, state_delay=2000
    )
    perfect_link = LhiLinkInfo.perfect(1000)
    nodes = {0: "node0", 1: "node1"}
    lhi_network = LhiNetworkInfo(
        nodes=nodes,
        links={frozenset([0, 1]): depolar_link, frozenset([1, 3]): perfect_link},
    )

    ehi_network = LhiConverter.network_to_ehi(lhi_network)
    expected_duration_0_1 = (10 / 0.5) + 2000
    expected_fidelty_0_1 = prob_max_mixed_to_fidelity(2, 0.2)

    ehi_link_0_1 = ehi_network.get_link(0, 1)
    assert ehi_link_0_1.duration == pytest.approx(expected_duration_0_1)
    assert ehi_link_0_1.fidelity == pytest.approx(expected_fidelty_0_1)

    ehi_link_1_3 = ehi_network.get_link(1, 3)
    assert ehi_link_1_3.duration == 1000
    assert ehi_link_1_3.fidelity == 1.0


if __name__ == "__main__":
    test_topology_to_ehi()
    test_topology_to_ehi_trapped_ion()
    test_qubit_to_ehi()
    test_gate_to_ehi()
    test_link_info_to_ehi_perfect()
    test_link_info_to_ehi_depolarise()
    test_network_to_ehi()



from __future__ import annotations

import os
import random
from dataclasses import dataclass
from typing import List

import netsquid as ns
from netqasm.lang.instr.flavour import NVFlavour

from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    LatenciesConfig,
    NtfConfig,
    NvParams,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.program import BatchResult, ProgramInput
from qoala.util.runner import run_two_node_app, run_two_node_app_separate_inputs


def create_procnode_cfg(name: str, id: int, num_qubits: int) -> ProcNodeConfig:
    return ProcNodeConfig(
        node_name=name,
        node_id=id,
        topology=TopologyConfig.from_nv_params(num_qubits=5, params=NvParams()),
        latencies=LatenciesConfig(qnos_instr_time=1000),
        ntf=NtfConfig.from_cls_name("NvNtf"),
        determ_sched=True,
    )


def load_program(path: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()
    return QoalaParser(text, flavour=NVFlavour()).parse()


@dataclass
class TeleportResult:
    alice_results: BatchResult
    bob_results: BatchResult


def run_teleport(num_iterations: int, different_inputs: bool = False) -> TeleportResult:
    ns.sim_reset()

    num_qubits = 3
    alice_id = 1
    bob_id = 0

    alice_node_cfg = create_procnode_cfg("alice", alice_id, num_qubits)
    bob_node_cfg = create_procnode_cfg("bob", bob_id, num_qubits)

    network_cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
        nodes=[alice_node_cfg, bob_node_cfg], link_duration=1000
    )

    alice_program = load_program("teleport_nv_alice.iqoala")
    bob_program = load_program("teleport_nv_bob.iqoala")

    if different_inputs:
        alice_inputs: List[ProgramInput] = []
        bob_inputs: List[ProgramInput] = []
        for _ in range(num_iterations):
            basis = random.randint(0, 5)
            alice_inputs.append(ProgramInput({"bob_id": bob_id, "state": basis}))
            bob_inputs.append(ProgramInput({"alice_id": alice_id, "state": basis}))

        app_result = run_two_node_app_separate_inputs(
            num_iterations=num_iterations,
            programs={"alice": alice_program, "bob": bob_program},
            program_inputs={"alice": alice_inputs, "bob": bob_inputs},
            network_cfg=network_cfg,
            linear=True,
        )
    else:
        # state = 5 -> teleport |1> state
        alice_input = ProgramInput({"bob_id": bob_id, "state": 5})
        # state = 4 -> measure in +Z, i.e. expect a "1" outcome
        bob_input = ProgramInput({"alice_id": alice_id, "state": 4})

        app_result = run_two_node_app(
            num_iterations=num_iterations,
            programs={"alice": alice_program, "bob": bob_program},
            program_inputs={"alice": alice_input, "bob": bob_input},
            network_cfg=network_cfg,
        )

    alice_result = app_result.batch_results["alice"]
    bob_result = app_result.batch_results["bob"]

    return TeleportResult(alice_result, bob_result)


def test_teleport():
    # LogManager.set_log_level("DEBUG")
    # LogManager.enable_task_logger(True)
    num_iterations = 10

    result = run_teleport(num_iterations=num_iterations)

    program_results = result.bob_results.results
    outcomes = [result.values["outcome"] for result in program_results]
    print(outcomes)
    assert all(outcome == 1 for outcome in outcomes)


def test_teleport_different_inputs():
    # LogManager.set_log_level("DEBUG")
    num_iterations = 10

    result = run_teleport(num_iterations=num_iterations, different_inputs=True)

    program_results = result.bob_results.results
    outcomes = [result.values["outcome"] for result in program_results]
    print(outcomes)
    assert all(outcome == 0 for outcome in outcomes)


if __name__ == "__main__":
    test_teleport()
    test_teleport_different_inputs()



from __future__ import annotations

import datetime
import json
import math
import os
import random
import time
from argparse import ArgumentParser
from curses.ascii import alt
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Dict, List, Tuple

import netsquid as ns

from qoala.lang.ehi import UnitModule
from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    ClassicalConnectionConfig,
    LatenciesConfig,
    NetworkScheduleConfig,
    NtfConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.program import BatchInfo, BatchResult, ProgramBatch, ProgramInput
from qoala.sim.build import build_network_from_config
from qoala.sim.network import ProcNodeNetwork
from qoala.util.logging import LogManager


def relative_to_cwd(file: str) -> str:
    return os.path.join(os.path.dirname(__file__), file)


def get_client_config(id: int) -> ProcNodeConfig:
    # client only needs 1 qubit
    return ProcNodeConfig(
        node_name=f"client",
        node_id=id,
        topology=TopologyConfig.perfect_config_uniform_default_params(20),
        latencies=LatenciesConfig(
            host_instr_time=500, host_peer_latency=30_000, qnos_instr_time=1000
        ),
        ntf=NtfConfig.from_cls_name("GenericNtf"),
    )


def get_server_config(id: int, num_qubits: int) -> ProcNodeConfig:
    return ProcNodeConfig(
        node_name="server",
        node_id=id,
        topology=TopologyConfig.perfect_config_uniform_default_params(num_qubits),
        latencies=LatenciesConfig(
            host_instr_time=500, host_peer_latency=30_000, qnos_instr_time=1000
        ),
        ntf=NtfConfig.from_cls_name("GenericNtf"),
    )


def create_network(
    server_cfg: ProcNodeConfig,
    client_cfg: ProcNodeConfig,
    num_iterations: int,
    qkd_num_iterations: int,
    cc: float,
    use_netschedule: bool,
    alternating: bool,
    bin_length: float,
) -> ProcNodeNetwork:
    node_cfgs = [server_cfg, client_cfg]
    network_cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
        nodes=node_cfgs, link_duration=1000
    )

    pattern: List[Tuple[int, int, int, int]] = []
    if alternating:
        # start index of PIDs per app
        bqc_start = 0
        qkd_start = num_iterations

        bqc_idx = 0
        qkd_idx = 0

        while bqc_idx < num_iterations or qkd_idx < qkd_num_iterations:
            if bqc_idx < num_iterations:
                pattern.append((0, bqc_start + bqc_idx, 1, bqc_start + bqc_idx))
            if qkd_idx < qkd_num_iterations:
                pattern.append((0, qkd_start + qkd_idx, 1, qkd_start + qkd_idx))
            bqc_idx += 1
            qkd_idx += 1
    else:
        for i in range(num_iterations + qkd_num_iterations):
            pattern.append((0, i, 1, i))

    if use_netschedule:
        network_cfg.netschedule = NetworkScheduleConfig(
            bin_length=bin_length,
            first_bin=0,
            bin_pattern=pattern,
            repeat_period=bin_length * (num_iterations + qkd_num_iterations),
        )

    cconns = [ClassicalConnectionConfig.from_nodes(0, 1, cc)]
    network_cfg.cconns = cconns
    return build_network_from_config(network_cfg)


@dataclass
class BqcQkdResult:
    client_results: BatchResult
    alice_results: BatchResult
    bob_results: BatchResult


@dataclass
class BqcParams:
    alpha: int
    beta: int
    theta1: int
    theta2: int
    dummy0: int
    dummy1: int


def load_program(path: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()
    return QoalaParser(text).parse()


def create_batch(
    program: QoalaProgram,
    inputs: List[ProgramInput],
    unit_module: UnitModule,
    num_iterations: int,
) -> BatchInfo:
    return BatchInfo(
        program=program,
        inputs=inputs,
        unit_module=unit_module,
        num_iterations=num_iterations,
        deadline=0,
    )


def create_client_inputs(
    bqc_params: BqcParams, server_id: int, num_iterations: int
) -> List[ProgramInput]:
    return [
        ProgramInput(
            {
                "server_id": server_id,
                "alpha": bqc_params.alpha,
                "beta": bqc_params.beta,
                "theta1": bqc_params.theta1,
                "theta2": bqc_params.theta2,
                "dummy0": bqc_params.dummy0,
                "dummy1": bqc_params.dummy1,
            }
        )
        for _ in range(num_iterations)
    ]


def run_bqc(
    bqc_params: BqcParams,
    num_iterations: int,
    qkd_num_iterations: int,
    qkd_num_pairs: int,
    linear: bool,
    cc: float,
    server_num_qubits: int,
    use_netschedule: bool,
    alternating: bool,
    bin_length: float,
) -> Tuple[BqcQkdResult, float]:
    ns.sim_reset()
    ns.set_qstate_formalism(ns.QFormalism.DM)
    seed = random.randint(0, 1000)
    ns.set_random_state(seed=seed)

    # Create network
    server_id = 0
    client_id = 1
    server_config = get_server_config(id=server_id, num_qubits=server_num_qubits)
    client_config = get_client_config(id=client_id)

    network = create_network(
        server_config,
        client_config,
        num_iterations,
        qkd_num_iterations,
        cc,
        use_netschedule,
        alternating,
        bin_length,
    )

    # Load programs for client node
    client_procnode = network.nodes[f"client"]

    # Load BQC client program
    client_program = load_program("vbqc_client.iqoala")
    client_inputs = create_client_inputs(bqc_params, server_id, num_iterations)
    client_unit_module = UnitModule.from_full_ehi(client_procnode.memmgr.get_ehi())
    client_batch_info = create_batch(
        program=client_program,
        inputs=client_inputs,
        unit_module=client_unit_module,
        num_iterations=num_iterations,
    )
    client_batch = client_procnode.submit_batch(client_batch_info)

    # Load QKD alice program
    alice_program = load_program("qkd_npairs_MD_alice.iqoala")
    alice_inputs = [
        ProgramInput({"bob_id": server_id, "N": qkd_num_pairs})
        for _ in range(qkd_num_iterations)
    ]
    alice_unit_module = UnitModule.from_full_ehi(client_procnode.memmgr.get_ehi())
    alice_batch_info = create_batch(
        program=alice_program,
        inputs=alice_inputs,
        unit_module=alice_unit_module,
        num_iterations=qkd_num_iterations,
    )
    alice_batch = client_procnode.submit_batch(alice_batch_info)

    # Load programs for server node
    server_procnode = network.nodes["server"]

    # Load BQC server program
    server_program = load_program("vbqc_server.iqoala")
    server_inputs = [
        ProgramInput({"client_id": client_id}) for _ in range(num_iterations)
    ]
    server_unit_module = UnitModule.from_full_ehi(server_procnode.memmgr.get_ehi())
    server_batch_info = create_batch(
        program=server_program,
        inputs=server_inputs,
        unit_module=server_unit_module,
        num_iterations=num_iterations,
    )
    server_batch = server_procnode.submit_batch(server_batch_info)

    # Load QKD bob program
    bob_program = load_program("qkd_npairs_MD_bob.iqoala")
    bob_inputs = [
        ProgramInput({"alice_id": client_id, "N": qkd_num_pairs})
        for _ in range(qkd_num_iterations)
    ]
    bob_unit_module = UnitModule.from_full_ehi(server_procnode.memmgr.get_ehi())
    bob_batch_info = create_batch(
        program=bob_program,
        inputs=bob_inputs,
        unit_module=bob_unit_module,
        num_iterations=qkd_num_iterations,
    )
    bob_batch = server_procnode.submit_batch(bob_batch_info)

    # Match PIDs
    remote_pids_for_server_node = {
        server_batch.batch_id: [inst.pid for inst in client_batch.instances],
        bob_batch.batch_id: [inst.pid for inst in bob_batch.instances],
    }
    remote_pids_for_client_node = {
        client_batch.batch_id: [inst.pid for inst in server_batch.instances],
        alice_batch.batch_id: [inst.pid for inst in alice_batch.instances],
    }
    # print(f"remote PIDs for server node: {remote_pids_for_server_node}")
    # print(f"remote PIDs for client node: {remote_pids_for_client_node}")
    client_procnode.initialize_processes(remote_pids_for_client_node, linear=linear)
    server_procnode.initialize_processes(remote_pids_for_server_node, linear=linear)

    network.start()
    start_time = ns.sim_time()
    ns.sim_run()
    end_time = ns.sim_time()
    makespan = end_time - start_time

    all_client_results = client_procnode.scheduler.get_batch_results()
    client_results = all_client_results[client_batch.batch_id]
    alice_results = all_client_results[alice_batch.batch_id]
    all_server_results = server_procnode.scheduler.get_batch_results()
    bob_results = all_server_results[bob_batch.batch_id]

    return BqcQkdResult(client_results, alice_results, bob_results), makespan


def bqc_computation(
    num_iterations: int,
    qkd_num_iterations: int,
    qkd_num_pairs: int,
    linear: bool,
    cc: float,
    server_num_qubits: int,
    use_netschedule: bool,
    alternating: bool,
    bin_length: float,
) -> Tuple[float, float]:
    ns.set_qstate_formalism(ns.qubits.qformalism.QFormalism.DM)

    bqc_params = BqcParams(alpha=8, beta=24, theta1=2, theta2=22, dummy0=0, dummy1=0)
    expected = 1

    bqc_result, makespan = run_bqc(
        bqc_params=bqc_params,
        num_iterations=num_iterations,
        qkd_num_iterations=qkd_num_iterations,
        qkd_num_pairs=qkd_num_pairs,
        linear=linear,
        cc=cc,
        server_num_qubits=server_num_qubits,
        use_netschedule=use_netschedule,
        alternating=alternating,
        bin_length=bin_length,
    )

    client_results = bqc_result.client_results
    alice_results = bqc_result.alice_results
    bob_results = bqc_result.bob_results

    m2s = [result.values["m2"] for result in client_results.results]
    correct_outcomes = len([m2 for m2 in m2s if m2 == expected])
    succ_prob = round(correct_outcomes / num_iterations, 2)

    alice_outcomes = [result.values["outcomes"][0] for result in alice_results.results]
    bob_outcomes = [result.values["outcomes"][0] for result in bob_results.results]
    # print(f"alice: {alice_outcomes}")
    # print(f"bob: {bob_outcomes}")
    assert alice_outcomes == bob_outcomes

    return makespan


@dataclass
class DataPoint:
    makespan_linear: float
    makespan_intlv_seq: float
    makespan_intlv_alt: float
    seq_improv: float
    alt_improv: float


@dataclass
class DataMeta:
    timestamp: str
    bqc_num_iterations: int
    qkd_num_iterations: int
    cc: float
    num_qubits: int
    bin_length: float
    sim_duration: float


@dataclass
class Data:
    meta: DataMeta
    data_points: List[DataPoint]


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--bqc_num_iterations", "-b", type=int, required=True)
    parser.add_argument("--qkd_num_iterations", "-q", type=int, required=True)

    args = parser.parse_args()
    bqc_num_iterations = args.bqc_num_iterations
    qkd_num_iterations = args.qkd_num_iterations

    qkd_num_pairs = 1
    cc = 1e5
    num_qubits = 5
    bin_length = 5e4

    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")

    start = time.time()

    linear_makespan = bqc_computation(
        bqc_num_iterations,
        qkd_num_iterations=qkd_num_iterations,
        qkd_num_pairs=qkd_num_pairs,
        linear=True,
        cc=cc,
        server_num_qubits=num_qubits,
        use_netschedule=False,
        alternating=False,
        bin_length=bin_length,
    )
    print(f"linear: makespan: {linear_makespan:_}")

    interleaved_seq_makespan = bqc_computation(
        bqc_num_iterations,
        qkd_num_iterations=qkd_num_iterations,
        qkd_num_pairs=qkd_num_pairs,
        linear=False,
        cc=cc,
        server_num_qubits=num_qubits,
        use_netschedule=True,
        alternating=False,
        bin_length=bin_length,
    )
    print(f"interleaved, sequential bins: makespan: {interleaved_seq_makespan:_}")

    interleaved_alt_makespan = bqc_computation(
        bqc_num_iterations,
        qkd_num_iterations=qkd_num_iterations,
        qkd_num_pairs=qkd_num_pairs,
        linear=False,
        cc=cc,
        server_num_qubits=num_qubits,
        use_netschedule=True,
        alternating=True,
        bin_length=bin_length,
    )
    print(f"interleaved, alternating bins: makespan: {interleaved_alt_makespan:_}")

    seq_improv = round(1 - interleaved_seq_makespan / linear_makespan, 3)
    alt_improv = round(1 - interleaved_alt_makespan / linear_makespan, 3)
    print(f"improvement for sequential netschedule: {seq_improv}")
    print(f"improvement for alternating netschedule: {alt_improv}")

    data_points = [
        DataPoint(
            makespan_linear=linear_makespan,
            makespan_intlv_seq=interleaved_seq_makespan,
            makespan_intlv_alt=interleaved_alt_makespan,
            seq_improv=seq_improv,
            alt_improv=alt_improv,
        )
    ]

    end = time.time()
    print(f"duration: {round(end - start, 2)} s")

    duration = round(end - start, 2)

    abs_dir = relative_to_cwd(f"data")
    Path(abs_dir).mkdir(parents=True, exist_ok=True)
    last_path = os.path.join(abs_dir, "LAST.json")
    timestamp_path = os.path.join(abs_dir, f"{timestamp}.json")

    meta = DataMeta(
        timestamp=timestamp,
        bqc_num_iterations=bqc_num_iterations,
        qkd_num_iterations=qkd_num_iterations,
        cc=cc,
        num_qubits=num_qubits,
        bin_length=bin_length,
        sim_duration=duration,
    )

    data = Data(meta=meta, data_points=data_points)
    json_data = asdict(data)

    with open(last_path, "w") as datafile:
        json.dump(json_data, datafile)
    with open(timestamp_path, "w") as datafile:
        json.dump(json_data, datafile)



from __future__ import annotations

from netsquid.protocols import Protocol

from qoala.lang.ehi import EhiNetworkInfo
from qoala.sim.memmgr import MemoryManager
from qoala.sim.netstack.netstackcomp import NetstackComponent
from qoala.sim.netstack.netstackinterface import NetstackInterface, NetstackLatencies
from qoala.sim.netstack.netstackprocessor import NetstackProcessor
from qoala.sim.qdevice import QDevice


class Netstack(Protocol):
    """NetSquid protocol representing the QNodeOS network stack."""

    def __init__(
        self,
        comp: NetstackComponent,
        ehi_network: EhiNetworkInfo,
        memmgr: MemoryManager,
        qdevice: QDevice,
        latencies: NetstackLatencies,
    ) -> None:
        """Network stack protocol constructor. Typically created indirectly through
        constructing a `Qnos` instance.

        :param comp: NetSquid component representing the network stack
        :param qnos: `Qnos` protocol that owns this protocol
        """
        super().__init__(name=f"{comp.name}_protocol")

        # References to objects.
        self._comp = comp
        self._ehi_network = ehi_network

        # Owned objects.
        self._interface = NetstackInterface(comp, ehi_network, qdevice, memmgr)
        self._processor = NetstackProcessor(self._interface, latencies)

    @property
    def qdevice(self) -> QDevice:
        return self._interface.qdevice

    @qdevice.setter
    def qdevice(self, qdevice: QDevice) -> None:
        self._interface._qdevice = qdevice

    @property
    def interface(self) -> NetstackInterface:
        return self._interface

    @interface.setter
    def interface(self, interface: NetstackInterface) -> None:
        self._interface = interface
        self._processor._interface = interface

    @property
    def processor(self) -> NetstackProcessor:
        return self._processor

    @processor.setter
    def processor(self, processor: NetstackProcessor) -> None:
        self._processor = processor

    def start(self) -> None:
        super().start()
        self._interface.start()

    def stop(self) -> None:
        self._interface.stop()
        super().stop()
